"""
RoPE ViT, adapted from
https://github.com/naver-ai/rope-vit/blob/main/deit/models_v2_rope.py
and
https://github.com/huggingface/pytorch-image-models/blob/main/timm/layers/pos_embed_sincos.py

Paper "Rotary Position Embedding for Vision Transformer", https://arxiv.org/abs/2403.13298

Changes from original:
* Implemented only axial RoPE (EVA style RoPE)
"""

# Reference license: Apache-2.0 and Apache-2.0

import math
from collections.abc import Callable
from functools import partial
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.layers import FFN
from birder.layers import LayerScale
from birder.layers import MultiHeadAttentionPool
from birder.layers import SwiGLU_FFN
from birder.layers.activations import get_activation_module
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenOmissionMixin
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenOmissionResultType
from birder.net.base import TokenRetentionResultType
from birder.net.vit import PatchEmbed
from birder.net.vit import adjust_position_embedding


def build_rotary_pos_embed(
    dim: int,
    temperature: float,
    grid_size: tuple[int, int],
    grid_indexing: str,
    grid_offset: int,
    pt_grid_size: Optional[tuple[int, int]],
) -> tuple[torch.Tensor, torch.Tensor]:
    assert dim % 4 == 0
    num_bands = dim // 4
    exp = torch.arange(0, num_bands, 1) / num_bands
    bands = 1.0 / (temperature**exp)

    if pt_grid_size is None:
        pt_grid_size = grid_size

    t = [(torch.arange(s) + grid_offset) / s * p for s, p in zip(grid_size, pt_grid_size)]
    grid = torch.stack(torch.meshgrid(t, indexing=grid_indexing), dim=-1)
    grid = grid.unsqueeze(-1)
    pos = grid * bands
    sin_emb = pos.sin()
    cos_emb = pos.cos()

    num_spatial_dim = grid_size[0] * grid_size[1]

    sin_emb = sin_emb.reshape(num_spatial_dim, -1).repeat_interleave(2, -1)
    cos_emb = cos_emb.reshape(num_spatial_dim, -1).repeat_interleave(2, -1)

    return (sin_emb, cos_emb)


def rotate_half(x: torch.Tensor) -> torch.Tensor:
    # Taken from: https://github.com/facebookresearch/capi/blob/main/model.py
    (x1, x2) = x.chunk(2, dim=-1)
    return torch.concat((-x2, x1), dim=-1)


def rotate_half_interleaved(x: torch.Tensor) -> torch.Tensor:
    return torch.stack([-x[..., 1::2], x[..., ::2]], dim=-1).reshape(x.size())


def apply_rotary_pos_embed(x: torch.Tensor, embed: torch.Tensor) -> torch.Tensor:
    (sin_emb, cos_emb) = embed.tensor_split(2, dim=-1)
    if cos_emb.ndim == 3:
        return x * cos_emb.unsqueeze(1).expand_as(x) + rotate_half(x) * sin_emb.unsqueeze(1).expand_as(x)

    return x * cos_emb + rotate_half(x) * sin_emb


def apply_interleaved_rotary_pos_embed(x: torch.Tensor, embed: torch.Tensor) -> torch.Tensor:
    (sin_emb, cos_emb) = embed.tensor_split(2, dim=-1)
    if cos_emb.ndim == 3:
        return x * cos_emb.unsqueeze(1).expand_as(x) + rotate_half_interleaved(x) * sin_emb.unsqueeze(1).expand_as(x)

    return x * cos_emb + rotate_half_interleaved(x) * sin_emb


class SequentialWithRope(nn.Sequential):
    def forward(self, x: torch.Tensor, rope: torch.Tensor) -> torch.Tensor:  # pylint: disable=arguments-differ
        for module in self:
            x = module(x, rope)

        return x


class RoPE(nn.Module):
    def __init__(
        self,
        dim: int,
        temperature: float,
        grid_size: tuple[int, int],
        grid_indexing: Literal["ij", "xy"],
        grid_offset: int,
        pt_grid_size: Optional[tuple[int, int]] = None,
        rope_rot_type: str = "standard",
    ) -> None:
        super().__init__()
        if rope_rot_type == "standard":
            self.apply_fn = apply_rotary_pos_embed
        elif rope_rot_type == "interleaved":
            self.apply_fn = apply_interleaved_rotary_pos_embed
        else:
            raise ValueError(f"Unknown rope_rot_type, got '{rope_rot_type}'")

        (sin_emb, cos_emb) = build_rotary_pos_embed(
            dim,
            temperature,
            grid_size=grid_size,
            grid_indexing=grid_indexing,
            grid_offset=grid_offset,
            pt_grid_size=pt_grid_size,
        )
        self.pos_embed = nn.Buffer(torch.concat((sin_emb, cos_emb), dim=-1), persistent=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.apply_fn(x, self.pos_embed)


class RoPEAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        attn_drop: float,
        proj_drop: float,
        num_special_tokens: int,
        rope_rot_type: str = "standard",
    ) -> None:
        super().__init__()
        assert dim % num_heads == 0, "dim should be divisible by num_heads"
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.scale = self.head_dim**-0.5
        self.num_special_tokens = num_special_tokens
        if rope_rot_type == "standard":
            self.apply_rot_fn = apply_rotary_pos_embed
        elif rope_rot_type == "interleaved":
            self.apply_rot_fn = apply_interleaved_rotary_pos_embed
        else:
            raise ValueError(f"Unknown rope_rot_type, got '{rope_rot_type}'")

        self.qkv = nn.Linear(dim, dim * 3)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: torch.Tensor, rope: torch.Tensor) -> torch.Tensor:
        (B, N, C) = x.size()
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, self.head_dim).permute(2, 0, 3, 1, 4)
        (q, k, v) = qkv.unbind(0)

        n = self.num_special_tokens
        q = torch.concat([q[:, :, :n, :], self.apply_rot_fn(q[:, :, n:, :], rope)], dim=2)
        k = torch.concat([k[:, :, :n, :], self.apply_rot_fn(k[:, :, n:, :], rope)], dim=2)

        x = F.scaled_dot_product_attention(  # pylint: disable=not-callable
            q, k, v, dropout_p=self.attn_drop.p if self.training else 0.0, scale=self.scale
        )

        x = x.transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class EncoderBlock(nn.Module):
    def __init__(
        self,
        num_heads: int,
        hidden_dim: int,
        mlp_dim: Optional[int],
        num_special_tokens: int,
        dropout: float,
        attention_dropout: float,
        drop_path: float,
        activation_layer: Callable[..., nn.Module],
        layer_scale_init_value: Optional[float] = None,
        norm_layer: Callable[..., nn.Module] = nn.LayerNorm,
        norm_layer_eps: float = 1e-6,
        mlp_layer: Callable[..., nn.Module] = FFN,
        rope_rot_type: str = "standard",
    ) -> None:
        super().__init__()

        if mlp_dim is None:
            mlp_dim = hidden_dim * 4

        # Attention block
        self.norm1 = norm_layer(hidden_dim, eps=norm_layer_eps)
        self.attn = RoPEAttention(
            hidden_dim,
            num_heads,
            attn_drop=attention_dropout,
            proj_drop=dropout,
            num_special_tokens=num_special_tokens,
            rope_rot_type=rope_rot_type,
        )
        if layer_scale_init_value is not None:
            self.layer_scale_1 = LayerScale(hidden_dim, layer_scale_init_value)
        else:
            self.layer_scale_1 = nn.Identity()

        # MLP block
        self.norm2 = norm_layer(hidden_dim, eps=norm_layer_eps)
        self.mlp = mlp_layer(hidden_dim, mlp_dim, act_layer=activation_layer, dropout=dropout)
        self.drop_path = StochasticDepth(drop_path, mode="row")
        if layer_scale_init_value is not None:
            self.layer_scale_2 = LayerScale(hidden_dim, layer_scale_init_value)
        else:
            self.layer_scale_2 = nn.Identity()

    def forward(self, x: torch.Tensor, rope: torch.Tensor) -> torch.Tensor:
        x = x + self.drop_path(self.layer_scale_1(self.attn(self.norm1(x), rope)))
        x = x + self.drop_path(self.layer_scale_2(self.mlp(self.norm2(x))))

        return x


class Encoder(nn.Module):
    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(
        self,
        num_layers: int,
        num_heads: int,
        hidden_dim: int,
        mlp_dim: int,
        num_special_tokens: int,
        dropout: float,
        attention_dropout: float,
        dpr: list[float],
        pre_norm: bool = False,
        activation_layer: Callable[..., nn.Module] = nn.GELU,
        layer_scale_init_value: Optional[float] = None,
        norm_layer: Callable[..., nn.Module] = nn.LayerNorm,
        norm_layer_eps: float = 1e-6,
        mlp_layer: Callable[..., nn.Module] = FFN,
        rope_rot_type: str = "standard",
    ) -> None:
        super().__init__()
        pre_layers = []
        if dropout > 0.0:
            pre_layers.append(nn.Dropout(dropout))
        if pre_norm is True:
            pre_layers.append(norm_layer(hidden_dim, eps=norm_layer_eps))

        self.pre_block = nn.Sequential(*pre_layers)

        layers = []
        for i in range(num_layers):
            layers.append(
                EncoderBlock(
                    num_heads,
                    hidden_dim,
                    mlp_dim,
                    num_special_tokens,
                    dropout,
                    attention_dropout,
                    dpr[i],
                    activation_layer=activation_layer,
                    layer_scale_init_value=layer_scale_init_value,
                    norm_layer=norm_layer,
                    norm_layer_eps=norm_layer_eps,
                    mlp_layer=mlp_layer,
                    rope_rot_type=rope_rot_type,
                )
            )

        self.block = SequentialWithRope(*layers)

    def forward(self, x: torch.Tensor, rope: torch.Tensor) -> torch.Tensor:
        x = self.pre_block(x)
        return self.block(x, rope)

    def forward_features(self, x: torch.Tensor, rope: torch.Tensor) -> list[torch.Tensor]:
        x = self.pre_block(x)

        xs = []
        for blk in self.block:
            x = blk(x, rope)
            xs.append(x)

        return xs


class MAEDecoderBlock(nn.Module):
    def __init__(
        self,
        num_heads: int,
        hidden_dim: int,
        num_special_tokens: int,
        activation_layer: Callable[..., nn.Module],
        grid_size: tuple[int, int],
        rope_grid_indexing: Literal["ij", "xy"],
        rope_grid_offset: int,
        rope_temperature: float,
        layer_scale_init_value: Optional[float] = None,
        norm_layer: Callable[..., nn.Module] = nn.LayerNorm,
        mlp_layer: Callable[..., nn.Module] = FFN,
        rope_rot_type: str = "standard",
    ) -> None:
        super().__init__()
        mlp_dim = hidden_dim * 4
        self.rope = RoPE(
            hidden_dim // num_heads,
            temperature=rope_temperature,
            grid_size=grid_size,
            grid_indexing=rope_grid_indexing,
            grid_offset=rope_grid_offset,
            rope_rot_type=rope_rot_type,
        )

        # Attention block
        self.norm1 = norm_layer(hidden_dim, eps=1e-6)
        self.attn = RoPEAttention(
            hidden_dim,
            num_heads,
            attn_drop=0.0,
            proj_drop=0.0,
            num_special_tokens=num_special_tokens,
            rope_rot_type=rope_rot_type,
        )
        if layer_scale_init_value is not None:
            self.layer_scale_1 = LayerScale(hidden_dim, layer_scale_init_value)
        else:
            self.layer_scale_1 = nn.Identity()

        # MLP block
        self.norm2 = norm_layer(hidden_dim, eps=1e-6)
        self.mlp = mlp_layer(hidden_dim, mlp_dim, act_layer=activation_layer, dropout=0.0)
        if layer_scale_init_value is not None:
            self.layer_scale_2 = LayerScale(hidden_dim, layer_scale_init_value)
        else:
            self.layer_scale_2 = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.layer_scale_1(self.attn(self.norm1(x), self.rope.pos_embed))
        x = x + self.layer_scale_2(self.mlp(self.norm2(x)))

        return x


# pylint: disable=invalid-name,too-many-instance-attributes
class RoPE_ViT(DetectorBackbone, PreTrainEncoder, MaskedTokenOmissionMixin, MaskedTokenRetentionMixin):
    block_group_regex = r"encoder\.block\.(\d+)"

    # pylint: disable=too-many-locals,too-many-branches,too-many-statements
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        image_size = self.size
        attention_dropout = 0.0
        dropout = 0.0
        pos_embed_special_tokens: bool = self.config.get("pos_embed_special_tokens", True)
        patch_size: int = self.config["patch_size"]
        num_layers: int = self.config["num_layers"]
        num_heads: int = self.config["num_heads"]
        hidden_dim: int = self.config["hidden_dim"]
        mlp_dim: int = self.config["mlp_dim"]
        layer_scale_init_value: Optional[float] = self.config.get("layer_scale_init_value", None)
        pre_norm: bool = self.config.get("pre_norm", False)
        post_norm: bool = self.config.get("post_norm", True)
        num_reg_tokens: int = self.config.get("num_reg_tokens", 0)
        class_token: bool = self.config.get("class_token", True)
        attn_pool_head: bool = self.config.get("attn_pool_head", False)
        attn_pool_num_heads: Optional[int] = self.config.get("attn_pool_num_heads", None)
        attn_pool_special_tokens: bool = self.config.get("attn_pool_special_tokens", False)
        norm_layer_type: str = self.config.get("norm_layer_type", "LayerNorm")
        norm_layer_eps: float = self.config.get("norm_layer_eps", 1e-6)
        mlp_layer_type: str = self.config.get("mlp_layer_type", "FFN")
        act_layer_type: Optional[str] = self.config.get("act_layer_type", None)  # Default according to mlp type
        rope_rot_type: Literal["standard", "interleaved"] = self.config.get("rope_rot_type", "standard")
        rope_grid_indexing: Literal["ij", "xy"] = self.config.get("rope_grid_indexing", "ij")
        rope_grid_offset: int = self.config.get("rope_grid_offset", 0)
        rope_temperature: float = self.config.get("rope_temperature", 100.0)
        pt_grid_size: Optional[tuple[int, int]] = self.config.get("pt_grid_size", None)
        drop_path_rate: float = self.config["drop_path_rate"]

        if norm_layer_type == "LayerNorm":
            norm_layer = nn.LayerNorm
        elif norm_layer_type == "RMSNorm":
            norm_layer = nn.RMSNorm
        else:
            raise ValueError(f"Unknown norm_layer_type '{norm_layer_type}'")

        if mlp_layer_type == "FFN":
            mlp_layer = FFN
            act_layer = nn.GELU
        elif mlp_layer_type == "SwiGLU_FFN":
            mlp_layer = SwiGLU_FFN
            act_layer = nn.SiLU
        else:
            raise ValueError(f"Unknown mlp_layer_type '{mlp_layer_type}'")

        if act_layer_type is not None:
            act_layer = get_activation_module(act_layer_type)

        torch._assert(image_size[0] % patch_size == 0, "Input shape indivisible by patch size!")
        torch._assert(image_size[1] % patch_size == 0, "Input shape indivisible by patch size!")
        torch._assert(hidden_dim % num_heads == 0, "Hidden dim indivisible by num heads!")
        self.pos_embed_special_tokens = pos_embed_special_tokens
        self.patch_size = patch_size
        self.num_layers = num_layers
        self.num_heads = num_heads
        self.hidden_dim = hidden_dim
        self.layer_scale_init_value = layer_scale_init_value
        self.num_reg_tokens = num_reg_tokens
        self.attn_pool_special_tokens = attn_pool_special_tokens
        self.norm_layer = norm_layer
        self.mlp_layer = mlp_layer
        self.act_layer = act_layer
        self.rope_rot_type = rope_rot_type
        self.rope_grid_indexing = rope_grid_indexing
        self.rope_grid_offset = rope_grid_offset
        self.rope_temperature = rope_temperature

        # Cast in case config was loaded from a json (no tuples),
        # TorchScript does not accept a list when tuple expected
        if isinstance(pt_grid_size, list):
            pt_grid_size = tuple(pt_grid_size)  # type: ignore[unreachable]

        self.pt_grid_size = pt_grid_size
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, num_layers)]  # Stochastic depth decay rule

        self.conv_proj = nn.Conv2d(
            self.input_channels,
            hidden_dim,
            kernel_size=(patch_size, patch_size),
            stride=(patch_size, patch_size),
            padding=(0, 0),
            bias=not pre_norm,
        )
        self.patch_embed = PatchEmbed()

        seq_length = (image_size[0] // patch_size) * (image_size[1] // patch_size)
        self.num_special_tokens = 0

        # Add a class token
        if class_token is True:
            self.class_token = nn.Parameter(torch.zeros(1, 1, hidden_dim))
            self.num_special_tokens += 1
            if pos_embed_special_tokens is True:
                seq_length += 1
        else:
            self.class_token = None

        # Add optional register tokens
        if self.num_reg_tokens > 0:
            self.reg_tokens = nn.Parameter(torch.zeros(1, self.num_reg_tokens, hidden_dim))
            self.num_special_tokens += self.num_reg_tokens
            if pos_embed_special_tokens is True:
                seq_length += self.num_reg_tokens
        else:
            self.reg_tokens = None

        # Add positional embedding
        self.pos_embedding = nn.Parameter(torch.empty(1, seq_length, hidden_dim).normal_(std=0.02))

        # RoPE
        self.rope = RoPE(
            hidden_dim // num_heads,
            temperature=self.rope_temperature,
            grid_size=(image_size[0] // patch_size, image_size[1] // patch_size),
            grid_indexing=rope_grid_indexing,
            grid_offset=rope_grid_offset,
            pt_grid_size=self.pt_grid_size,
            rope_rot_type=rope_rot_type,
        )

        # Encoder
        self.encoder = Encoder(
            num_layers,
            num_heads,
            hidden_dim,
            mlp_dim,
            self.num_special_tokens,
            dropout,
            attention_dropout,
            dpr,
            pre_norm=pre_norm,
            activation_layer=act_layer,
            layer_scale_init_value=layer_scale_init_value,
            norm_layer=norm_layer,
            norm_layer_eps=norm_layer_eps,
            mlp_layer=mlp_layer,
            rope_rot_type=rope_rot_type,
        )

        if post_norm is True:
            self.norm = norm_layer(hidden_dim, eps=norm_layer_eps)
        else:
            self.norm = nn.Identity()

        if attn_pool_head is False:
            self.attn_pool = None
        else:
            if attn_pool_num_heads is None:
                attn_pool_num_heads = num_heads

            self.attn_pool = MultiHeadAttentionPool(hidden_dim, attn_pool_num_heads, mlp_dim, qkv_bias=True)

        self.return_stages = ["neck"]  # Actually meaningless, just for completeness
        self.return_channels = [hidden_dim]
        self.embedding_size = hidden_dim
        self.classifier = self.create_classifier()

        self.max_stride = patch_size
        self.stem_stride = patch_size
        self.stem_width = hidden_dim
        self.encoding_size = hidden_dim
        self.decoder_block = partial(
            MAEDecoderBlock,
            16,
            num_special_tokens=self.num_special_tokens,
            activation_layer=act_layer,
            grid_size=(image_size[0] // patch_size, image_size[1] // patch_size),
            rope_grid_indexing=rope_grid_indexing,
            rope_grid_offset=rope_grid_offset,
            rope_temperature=rope_temperature,
            layer_scale_init_value=layer_scale_init_value,
            norm_layer=norm_layer,
            mlp_layer=mlp_layer,
            rope_rot_type=rope_rot_type,
        )

        # Weight initialization
        if isinstance(self.conv_proj, nn.Conv2d):
            # Init the patchify stem
            fan_in = self.conv_proj.in_channels * self.conv_proj.kernel_size[0] * self.conv_proj.kernel_size[1]
            nn.init.trunc_normal_(self.conv_proj.weight, std=math.sqrt(1 / fan_in))
            if self.conv_proj.bias is not None:
                nn.init.zeros_(self.conv_proj.bias)

        if isinstance(self.classifier, nn.Linear):
            nn.init.zeros_(self.classifier.weight)
            nn.init.zeros_(self.classifier.bias)

    def _get_pos_embed(self, H: int, W: int) -> torch.Tensor:
        if self.dynamic_size is False:
            return self.pos_embedding

        if H == self.size[0] and W == self.size[1]:
            return self.pos_embedding

        return adjust_position_embedding(
            self.pos_embedding,
            (self.size[0] // self.patch_size, self.size[1] // self.patch_size),
            (H // self.patch_size, W // self.patch_size),
            self.num_special_tokens if self.pos_embed_special_tokens is True else 0,
            antialias=False,
        )

    def _get_rope_embed(self, H: int, W: int) -> torch.Tensor:
        if self.dynamic_size is False:
            return self.rope.pos_embed

        if H == self.size[0] and W == self.size[1]:
            return self.rope.pos_embed

        return torch.concat(
            build_rotary_pos_embed(
                self.hidden_dim // self.num_heads,
                self.rope_temperature,
                grid_size=(H // self.patch_size, W // self.patch_size),
                grid_indexing=self.rope_grid_indexing,
                grid_offset=self.rope_grid_offset,
                pt_grid_size=self.pt_grid_size,
            ),
            dim=-1,
        ).to(self.rope.pos_embed.device)

    def freeze(self, freeze_classifier: bool = True, unfreeze_features: bool = False) -> None:
        for param in self.parameters():
            param.requires_grad = False

        if freeze_classifier is False:
            for param in self.classifier.parameters():
                param.requires_grad = True

        if unfreeze_features is True:
            if self.attn_pool is not None:
                for param in self.attn_pool.parameters():
                    param.requires_grad = True

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        (H, W) = x.shape[-2:]
        x = self.conv_proj(x)
        x = self.patch_embed(x)

        if self.pos_embed_special_tokens is False:
            x = x + self._get_pos_embed(H, W)

        # Expand the class token to the full batch
        if self.class_token is not None:
            batch_class_token = self.class_token.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_class_token, x], dim=1)

        # Expand the register tokens to the full batch
        if self.reg_tokens is not None:
            batch_reg_tokens = self.reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        if self.pos_embed_special_tokens is True:
            x = x + self._get_pos_embed(H, W)

        x = self.encoder(x, self._get_rope_embed(H, W))
        x = self.norm(x)

        x = x[:, self.num_special_tokens :]
        x = x.permute(0, 2, 1)
        (B, C, _) = x.size()
        x = x.reshape(B, C, H // self.patch_size, W // self.patch_size)

        return {self.return_stages[0]: x}

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.conv_proj.parameters():
            param.requires_grad = False

        self.pos_embedding.requires_grad = False

        for idx, module in enumerate(self.encoder.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_omission(
        self,
        x: torch.Tensor,
        ids_keep: Optional[torch.Tensor] = None,
        return_all_features: bool = False,
        return_keys: Literal["all", "tokens", "embedding"] = "tokens",
    ) -> TokenOmissionResultType:
        (H, W) = x.shape[-2:]

        # Reshape and permute the input tensor
        x = self.conv_proj(x)
        x = self.patch_embed(x)

        # Add pos embedding without special tokens
        pos_embedding = self._get_pos_embed(H, W)
        if self.pos_embed_special_tokens is True:
            x = x + pos_embedding[:, self.num_special_tokens :, :]
        else:
            x = x + pos_embedding

        # Mask tokens
        if ids_keep is not None:
            x = torch.gather(x, dim=1, index=ids_keep.unsqueeze(-1).repeat(1, 1, x.size(2)))

            rope_dim = self.rope.pos_embed.size(1)
            rope = self.rope.pos_embed.unsqueeze(0).repeat(x.size(0), 1, 1)
            rope_masked = torch.gather(rope, dim=1, index=ids_keep.unsqueeze(-1).repeat(1, 1, rope_dim))
        else:
            rope_masked = self.rope.pos_embed

        # Append class and register tokens
        if self.class_token is not None:
            if self.pos_embed_special_tokens is True:
                cls_token = self.class_token + pos_embedding[:, self.num_reg_tokens : self.num_reg_tokens + 1, :]
            else:
                cls_token = self.class_token

            batch_class_token = cls_token.expand(x.shape[0], -1, -1)
            x = torch.concat((batch_class_token, x), dim=1)

        if self.reg_tokens is not None:
            if self.pos_embed_special_tokens is True:
                reg_tokens = self.reg_tokens + pos_embedding[:, 0 : self.num_reg_tokens, :]
            else:
                reg_tokens = self.reg_tokens

            batch_reg_tokens = reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        # Apply transformer
        if return_all_features is True:
            xs = self.encoder.forward_features(x, rope_masked)
            xs[-1] = self.norm(xs[-1])
            x = torch.stack(xs, dim=-1)
        else:
            x = self.encoder(x, rope_masked)
            x = self.norm(x)

        result: TokenOmissionResultType = {}
        if return_keys in ("all", "tokens"):
            result["tokens"] = x

        if return_keys in ("all", "embedding"):
            if return_all_features is True:
                x = x[..., -1]

            if self.attn_pool is not None:
                if self.attn_pool_special_tokens is False:
                    x = x[:, self.num_special_tokens :]

                x = self.attn_pool(x)
                result["embedding"] = x[:, 0]
            elif self.class_token is None:
                x = x[:, self.num_special_tokens :]
                result["embedding"] = x.mean(dim=1)
            else:
                result["embedding"] = x[:, self.num_reg_tokens]

        return result

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        (H, W) = x.shape[-2:]

        x = self.conv_proj(x)
        x = mask_tensor(x, mask, mask_token=mask_token, patch_factor=self.max_stride // self.stem_stride)

        # Reshape and permute the input tensor
        x = self.patch_embed(x)

        if self.pos_embed_special_tokens is False:
            x = x + self._get_pos_embed(H, W)

        # Expand the class token to the full batch
        if self.class_token is not None:
            batch_class_token = self.class_token.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_class_token, x], dim=1)

        # Expand the register tokens to the full batch
        if self.reg_tokens is not None:
            batch_reg_tokens = self.reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        if self.pos_embed_special_tokens is True:
            x = x + self._get_pos_embed(H, W)

        x = self.encoder(x, self._get_rope_embed(H, W))
        x = self.norm(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            features = x[:, self.num_special_tokens :]
            features = features.permute(0, 2, 1)
            (B, C, _) = features.size()
            features = features.reshape(B, C, H // self.patch_size, W // self.patch_size)
            result["features"] = features

        if return_keys in ("all", "embedding"):
            if self.attn_pool is not None:
                if self.attn_pool_special_tokens is False:
                    x = x[:, self.num_special_tokens :]

                x = self.attn_pool(x)
                result["embedding"] = x[:, 0]
            elif self.class_token is None:
                x = x[:, self.num_special_tokens :]
                result["embedding"] = x.mean(dim=1)
            else:
                result["embedding"] = x[:, self.num_reg_tokens]

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        (H, W) = x.shape[-2:]

        # Reshape and permute the input tensor
        x = self.conv_proj(x)
        x = self.patch_embed(x)

        if self.pos_embed_special_tokens is False:
            x = x + self._get_pos_embed(H, W)

        # Expand the class token to the full batch
        if self.class_token is not None:
            batch_class_token = self.class_token.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_class_token, x], dim=1)

        # Expand the register tokens to the full batch
        if self.reg_tokens is not None:
            batch_reg_tokens = self.reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        if self.pos_embed_special_tokens is True:
            x = x + self._get_pos_embed(H, W)

        x = self.encoder(x, self._get_rope_embed(H, W))
        x = self.norm(x)

        return x

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)

        if self.attn_pool is not None:
            if self.attn_pool_special_tokens is False:
                x = x[:, self.num_special_tokens :]

            x = self.attn_pool(x)
            return x[:, 0]

        if self.class_token is None:
            x = x[:, self.num_special_tokens :]
            return x.mean(dim=1)

        # Classifier "token" as used by standard language architectures
        return x[:, self.num_reg_tokens]

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        assert new_size[0] % self.patch_size == 0, "Input shape indivisible by patch size!"
        assert new_size[1] % self.patch_size == 0, "Input shape indivisible by patch size!"

        old_size = self.size
        super().adjust_size(new_size)

        if self.pos_embed_special_tokens is True:
            num_prefix_tokens = self.num_special_tokens
        else:
            num_prefix_tokens = 0

        # Add back class tokens
        self.pos_embedding = nn.Parameter(
            adjust_position_embedding(
                self.pos_embedding,
                (old_size[0] // self.patch_size, old_size[1] // self.patch_size),
                (new_size[0] // self.patch_size, new_size[1] // self.patch_size),
                num_prefix_tokens,
            )
        )

        # Adjust RoPE
        self.rope = RoPE(
            self.hidden_dim // self.num_heads,
            temperature=self.rope_temperature,
            grid_size=(new_size[0] // self.patch_size, new_size[1] // self.patch_size),
            grid_indexing=self.rope_grid_indexing,
            grid_offset=self.rope_grid_offset,
            pt_grid_size=self.pt_grid_size,
            rope_rot_type=self.rope_rot_type,
        )

        # Define adjusted decoder block
        self.decoder_block = partial(
            MAEDecoderBlock,
            16,
            num_special_tokens=self.num_special_tokens,
            activation_layer=self.act_layer,
            grid_size=(new_size[0] // self.patch_size, new_size[1] // self.patch_size),
            rope_grid_indexing=self.rope_grid_indexing,
            rope_grid_offset=self.rope_grid_offset,
            rope_temperature=self.rope_temperature,
            layer_scale_init_value=self.layer_scale_init_value,
            norm_layer=self.norm_layer,
            mlp_layer=self.mlp_layer,
            rope_rot_type=self.rope_rot_type,
        )


# Vision Transformer Model Naming Convention
# ==========================================
#
# Model names follow a structured pattern to encode architectural choices:
# [rope_]vit_[reg{N}_][size][patch_size][_components][_pooling][_c{N}]
#
# Core Components:
# - rope_       : Rotary Position Embedding (RoPE) enabled
# - rope_i_     : Rotary Position Embedding (RoPE) enabled with interleaved rotation - implies different temp, indexing
# - vit_        : Vision Transformer base architecture
# - reg{N}_     : Register tokens (N = number of register tokens, e.g., reg4, reg8)
# - size        : Model size (s=small, b=base, l=large, or specific like so150m)
# - patch_size  : Patch size (e.g., 14, 16, 32 for 14x14, 16x16, 32x32 patches)
#
# Optional Components:
#     Position Embeddings:
#     - nps         : No Position embedding on Special tokens
#
#     Normalization:
#     - rms         : RMSNorm (instead of LayerNorm)
#     - pn          : Pre-Norm (layer norm before the encoder) - implies different norm eps
#     - npn         : No Post Norm (disables post-normalization layer)
#
#     Feed-Forward Network:
#     - swiglu      : SwiGLU FFN layer type (instead of standard FFN)
#
#     Activation:
#     - quick_gelu  : QuickGELU activation type
#     - ...
#
#     Regularization:
#     - ls          : Layer Scaling applied
#
#     Pooling/Reduction:
#     - avg         : Average pooling for sequence reduction
#     - ap          : Attention Pooling for sequence reduction
#
#     Custom Variants:
#     - c{N}        : Custom variant (N = version number) for models with fine-grained or non-standard
#                     modifications not fully reflected in the name

registry.register_model_config(
    "rope_vit_s32",
    RoPE_ViT,
    config={
        "patch_size": 32,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "rope_vit_s16",
    RoPE_ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "rope_i_vit_s16_pn_ap_c1",  # For PE Core - https://arxiv.org/abs/2504.13181
    RoPE_ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "pre_norm": True,
        "attn_pool_head": True,
        "attn_pool_num_heads": 8,
        "attn_pool_special_tokens": True,
        "norm_layer_eps": 1e-5,
        "rope_rot_type": "interleaved",
        "rope_grid_indexing": "xy",
        "rope_grid_offset": 1,
        "rope_temperature": 10000.0,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "rope_vit_s14",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "rope_vit_m32",
    RoPE_ViT,
    config={
        "patch_size": 32,
        "num_layers": 12,
        "num_heads": 8,
        "hidden_dim": 512,
        "mlp_dim": 2048,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "rope_vit_m16",
    RoPE_ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 8,
        "hidden_dim": 512,
        "mlp_dim": 2048,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "rope_vit_m14",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 8,
        "hidden_dim": 512,
        "mlp_dim": 2048,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "rope_vit_b32",
    RoPE_ViT,
    config={
        "patch_size": 32,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "rope_vit_b16",
    RoPE_ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_i_vit_b16_pn_ap_c1",  # For PE Core - https://arxiv.org/abs/2504.13181
    RoPE_ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "pre_norm": True,
        "attn_pool_head": True,
        "attn_pool_num_heads": 8,
        "attn_pool_special_tokens": True,
        "norm_layer_eps": 1e-5,
        "rope_rot_type": "interleaved",
        "rope_grid_indexing": "xy",
        "rope_grid_offset": 1,
        "rope_temperature": 10000.0,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_b14",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_l32",
    RoPE_ViT,
    config={
        "patch_size": 32,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_l16",
    RoPE_ViT,
    config={
        "patch_size": 16,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_l14",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_i_vit_l14_pn_ap_c1",  # For PE Core - https://arxiv.org/abs/2504.13181
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "pre_norm": True,
        "attn_pool_head": True,
        "attn_pool_num_heads": 8,
        "attn_pool_special_tokens": True,
        "norm_layer_eps": 1e-5,
        "rope_rot_type": "interleaved",
        "rope_grid_indexing": "xy",
        "rope_grid_offset": 1,
        "rope_temperature": 10000.0,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_h16",
    RoPE_ViT,
    config={
        "patch_size": 16,
        "num_layers": 32,
        "num_heads": 16,
        "hidden_dim": 1280,
        "mlp_dim": 5120,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_h14",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 32,
        "num_heads": 16,
        "hidden_dim": 1280,
        "mlp_dim": 5120,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(  # From "Scaling Vision Transformers"
    "rope_vit_g14",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 40,
        "num_heads": 16,
        "hidden_dim": 1408,
        "mlp_dim": 6144,
        "drop_path_rate": 0.1,
    },
)

# With registers
registry.register_model_config(
    "rope_vit_reg1_s32",
    RoPE_ViT,
    config={
        "patch_size": 32,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "num_reg_tokens": 1,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "rope_vit_reg1_s16",
    RoPE_ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "num_reg_tokens": 1,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "rope_i_vit_reg1_s16_pn_npn_avg_c1",  # For PE Spatial - https://arxiv.org/abs/2504.13181
    RoPE_ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "num_reg_tokens": 1,
        "class_token": False,
        "pre_norm": True,
        "post_norm": False,
        "norm_layer_eps": 1e-5,
        "rope_rot_type": "interleaved",
        "rope_grid_indexing": "xy",
        "rope_grid_offset": 1,
        "rope_temperature": 10000.0,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "rope_vit_reg1_s14",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "num_reg_tokens": 1,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "rope_vit_reg4_m32",
    RoPE_ViT,
    config={
        "patch_size": 32,
        "num_layers": 12,
        "num_heads": 8,
        "hidden_dim": 512,
        "mlp_dim": 2048,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "rope_vit_reg4_m16",
    RoPE_ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 8,
        "hidden_dim": 512,
        "mlp_dim": 2048,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "rope_vit_reg4_m16_rms_avg",
    RoPE_ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 8,
        "hidden_dim": 512,
        "mlp_dim": 2048,
        "num_reg_tokens": 4,
        "class_token": False,
        "norm_layer_type": "RMSNorm",
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "rope_vit_reg4_m14",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 8,
        "hidden_dim": 512,
        "mlp_dim": 2048,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "rope_vit_reg4_b32",
    RoPE_ViT,
    config={
        "patch_size": 32,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "rope_vit_reg4_b16",
    RoPE_ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_reg4_b14",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_reg8_nps_b14_ap",
    RoPE_ViT,
    config={
        "pos_embed_special_tokens": False,
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "num_reg_tokens": 8,
        "class_token": False,
        "attn_pool_head": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_reg4_l32",
    RoPE_ViT,
    config={
        "patch_size": 32,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_reg4_l16",
    RoPE_ViT,
    config={
        "patch_size": 16,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_reg4_l14",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_reg8_l14_ap",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "num_reg_tokens": 8,
        "class_token": False,
        "attn_pool_head": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_reg8_l14_rms_ap",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "num_reg_tokens": 8,
        "class_token": False,
        "attn_pool_head": True,
        "norm_layer_type": "RMSNorm",
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_reg4_h16",
    RoPE_ViT,
    config={
        "patch_size": 16,
        "num_layers": 32,
        "num_heads": 16,
        "hidden_dim": 1280,
        "mlp_dim": 5120,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_reg4_h14",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 32,
        "num_heads": 16,
        "hidden_dim": 1280,
        "mlp_dim": 5120,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(  # From "Scaling Vision Transformers"
    "rope_vit_reg4_g14",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 40,
        "num_heads": 16,
        "hidden_dim": 1408,
        "mlp_dim": 6144,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.1,
    },
)

# Shape-optimized vision transformer (SoViT)
registry.register_model_config(
    "rope_vit_so150m_p14_ap",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 18,
        "num_heads": 16,
        "hidden_dim": 896,  # Changed from 880 for RoPE divisibility
        "mlp_dim": 2320,
        "class_token": False,
        "attn_pool_head": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_so400m_p14_ap",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 27,
        "num_heads": 16,
        "hidden_dim": 1152,
        "mlp_dim": 4304,
        "class_token": False,
        "attn_pool_head": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_reg4_so150m_p14_ap",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 18,
        "num_heads": 16,
        "hidden_dim": 896,  # Changed from 880 for RoPE divisibility
        "mlp_dim": 2320,
        "num_reg_tokens": 4,
        "class_token": False,
        "attn_pool_head": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_reg8_so150m_p14_swiglu_rms_avg",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 18,
        "num_heads": 16,
        "hidden_dim": 896,  # Changed from 880 for RoPE divisibility
        "mlp_dim": 2320,
        "num_reg_tokens": 8,
        "class_token": False,
        "norm_layer_type": "RMSNorm",
        "mlp_layer_type": "SwiGLU_FFN",
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_reg8_so150m_p14_swiglu_rms_ap",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 18,
        "num_heads": 16,
        "hidden_dim": 896,  # Changed from 880 for RoPE divisibility
        "mlp_dim": 2320,
        "num_reg_tokens": 8,
        "class_token": False,
        "attn_pool_head": True,
        "norm_layer_type": "RMSNorm",
        "mlp_layer_type": "SwiGLU_FFN",
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_reg8_so150m_p14_ap",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 18,
        "num_heads": 16,
        "hidden_dim": 896,  # Changed from 880 for RoPE divisibility
        "mlp_dim": 2320,
        "num_reg_tokens": 8,
        "class_token": False,
        "attn_pool_head": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_reg4_so400m_p14_ap",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 27,
        "num_heads": 16,
        "hidden_dim": 1152,
        "mlp_dim": 4304,
        "num_reg_tokens": 4,
        "class_token": False,
        "attn_pool_head": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "rope_vit_reg8_so400m_p14_ap",
    RoPE_ViT,
    config={
        "patch_size": 14,
        "num_layers": 27,
        "num_heads": 16,
        "hidden_dim": 1152,
        "mlp_dim": 4304,
        "num_reg_tokens": 8,
        "class_token": False,
        "attn_pool_head": True,
        "drop_path_rate": 0.1,
    },
)

registry.register_weights(
    "rope_vit_reg4_b14_capi",
    {
        "url": "https://huggingface.co/birder-project/rope_vit_reg4_b14_capi/resolve/main",
        "description": (
            "RoPE ViT b14 image encoder pre-trained using CAPI. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 327.0,
                "sha256": "175378d81734649567bfe82aac9557f9b0bf48dbd562f26e338b1958fa057472",
            }
        },
        "net": {"network": "rope_vit_reg4_b14", "tag": "capi"},
    },
)
registry.register_weights(
    "rope_vit_reg4_b14_capi-places365",
    {
        "url": "https://huggingface.co/birder-project/rope_vit_reg4_b14_capi-places365/resolve/main",
        "description": "RoPE ViT b14 model pre-trained using CAPI, then fine-tuned on the Places365 dataset",
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 328.1,
                "sha256": "4d3ef700eb0d454c9406e9b5c11f70106b46ed4a6ca24c1d89a60097ad78ea9a",
            }
        },
        "net": {"network": "rope_vit_reg4_b14", "tag": "capi-places365"},
    },
)
registry.register_weights(
    "rope_vit_reg4_b14_capi-inat21-224px",
    {
        "url": "https://huggingface.co/birder-project/rope_vit_reg4_b14_capi-inat21/resolve/main",
        "description": "RoPE ViT b14 model pre-trained using CAPI, then fine-tuned on the iNaturalist 2021 dataset",
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 357.2,
                "sha256": "fb98a4f29a1c6e552a4e22eaf614b0f2d2adedefe2d510fa7e69309208dc0f9f",
            }
        },
        "net": {"network": "rope_vit_reg4_b14", "tag": "capi-inat21-224px"},
    },
)
registry.register_weights(
    "rope_vit_reg4_b14_capi-inat21",
    {
        "url": "https://huggingface.co/birder-project/rope_vit_reg4_b14_capi-inat21/resolve/main",
        "description": "RoPE ViT b14 model pre-trained using CAPI, then fine-tuned on the iNaturalist 2021 dataset",
        "resolution": (336, 336),
        "formats": {
            "pt": {
                "file_size": 358.2,
                "sha256": "25befb5a460cc80a5a7961db61e747916461bf6967f3d39d9294ee474bd31304",
            }
        },
        "net": {"network": "rope_vit_reg4_b14", "tag": "capi-inat21"},
    },
)
registry.register_weights(
    "rope_vit_reg4_b14_capi-imagenet21k",
    {
        "url": "https://huggingface.co/birder-project/rope_vit_reg4_b14_capi-imagenet21k/resolve/main",
        "description": "RoPE ViT b14 model pre-trained using CAPI, then fine-tuned on the ImageNet-21K dataset",
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 383.7,
                "sha256": "1325f35f0d8dff3270d6ce645f81865e9b8de7bacf17f94a9f5e2ef0cd66f56d",
            }
        },
        "net": {"network": "rope_vit_reg4_b14", "tag": "capi-imagenet21k"},
    },
)

# Perception Encoder: The best visual embeddings are not at the output of the network, by Meta FAIR
# https://arxiv.org/abs/2504.13181
registry.register_weights(
    "rope_i_vit_s16_pn_ap_c1_pe-core",
    {
        "url": "https://huggingface.co/birder-project/rope_i_vit_s16_pn_ap_c1_pe-core/resolve/main",
        "description": (
            "ViT s16 image encoder pre-trained by Meta FAIR using CLIP. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (384, 384),
        "formats": {
            "pt": {
                "file_size": 90.0,
                "sha256": "e4429b0bafb9f827698dde73c882c70deb994329ea0dd169f68e76ad256bbb74",
            },
        },
        "net": {"network": "rope_i_vit_s16_pn_ap_c1", "tag": "pe-core"},
    },
)
registry.register_weights(
    "rope_i_vit_reg1_s16_pn_npn_avg_c1_pe-spatial",
    {
        "url": "https://huggingface.co/birder-project/rope_i_vit_reg1_s16_pn_npn_avg_c1_pe-spatial/resolve/main",
        "description": (
            "ViT s16 image encoder pre-trained by Meta FAIR using CLIP. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (512, 512),
        "formats": {
            "pt": {
                "file_size": 83.9,
                "sha256": "4e65e500f2a7d2b11fc28aaa0b1ad4921692780507de014ebc5659e757327fde",
            },
        },
        "net": {"network": "rope_i_vit_reg1_s16_pn_npn_avg_c1", "tag": "pe-spatial"},
    },
)
registry.register_weights(
    "rope_i_vit_b16_pn_ap_c1_pe-core",
    {
        "url": "https://huggingface.co/birder-project/rope_i_vit_b16_pn_ap_c1_pe-core/resolve/main",
        "description": (
            "ViT b16 image encoder pre-trained by Meta FAIR using CLIP. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 354.4,
                "sha256": "d1c1ba1e8c841f495ff3c0e5e6963a39c8d1ae07dea30d3b82422017a4062d97",
            },
        },
        "net": {"network": "rope_i_vit_b16_pn_ap_c1", "tag": "pe-core"},
    },
)
registry.register_weights(
    "rope_i_vit_l14_pn_ap_c1_pe-core",
    {
        "url": "https://huggingface.co/birder-project/rope_i_vit_l14_pn_ap_c1_pe-core/resolve/main",
        "description": (
            "ViT l14 image encoder pre-trained by Meta FAIR using CLIP. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (336, 336),
        "formats": {
            "pt": {
                "file_size": 1206.0,
                "sha256": "26c2188116cb254d2870c23cc3ab7d60d9ee0606c803b8dbe359e5716498b5c4",
            },
        },
        "net": {"network": "rope_i_vit_l14_pn_ap_c1", "tag": "pe-core"},
    },
)



import argparse

from birder.common import cli
from birder.tools import adversarial
from birder.tools import avg_model
from birder.tools import convert_model
from birder.tools import det_results
from birder.tools import download_model
from birder.tools import ensemble_model
from birder.tools import introspection
from birder.tools import labelme_to_coco
from birder.tools import list_models
from birder.tools import model_info
from birder.tools import pack
from birder.tools import quantize_model
from birder.tools import results
from birder.tools import show_det_iterator
from birder.tools import show_iterator
from birder.tools import similarity
from birder.tools import stats
from birder.tools import verify_coco
from birder.tools import verify_directory
from birder.tools import voc_to_coco


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="python -m birder.tools",
        allow_abbrev=False,
        description="Tool to run auxiliary commands",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools adversarial --method fgsm -n swin_transformer_v1_s -e 0 "
            "--image 'data/training/Mallard/000112.jpeg'\n"
            "python -m birder.tools avg-model --network resnet_v2_50 --epochs 95 95 100\n"
            "python -m birder.tools convert-model --network convnext_v2_base --epoch 0 --pt2\n"
            "python -m birder.tools det-results "
            "results/faster_rcnn_coco_csp_resnet_50_imagenet1k_91_e0_640px_5000.json --print\n"
            "python -m birder.tools download-model mobilenet_v3_large_1_0\n"
            "python -m birder.tools ensemble-model --network convnext_v2_4_0 focalnet_3_0 --pts\n"
            "python -m birder.tools introspection --method gradcam --network efficientnet_v2_m "
            "--epoch 200 --image 'data/validation/Mallard/000003.jpeg'\n"
            "python -m birder.tools labelme-to-coco data/detection_data\n"
            "python -m birder.tools list-models --pretrained\n"
            "python -m birder.tools model-info -n deit_s16 -t intermediate -e 0\n"
            "python -m birder.tools pack data/training\n"
            "python -m birder.tools quantize-model -n convnext_v2_base -e 0 --qbackend x86\n"
            "python -m birder.tools results results/inception_resnet_v2_105_e100_3150.csv --print --pr-curve\n"
            "python -m birder.tools show-det-iterator --mode inference --size 640 --batch\n"
            "python -m birder.tools show-iterator --mode training --size 256 320 --aug-level 5\n"
            "python -m birder.tools similarity -n efficientnet_v2_l -e 0 --limit 15 data/*/*crane\n"
            "python -m birder.tools stats --class-graph\n"
            "python -m birder.tools verify-coco --coco-json-path "
            "~/Datasets/Objects365-2020/train/zhiyuan_objv2_train.json --data-path ~/Datasets/Objects365-2020/train\n"
            "python -m birder.tools verify-directory data/testing\n"
            "python -m birder.tools voc-to-coco --class-file public_datasets_metadata/voc-classes.txt "
            "--ann-dir ~/Datasets/VOC2012/Annotations ~/Datasets/VOC2012/JPEGImages\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparsers = parser.add_subparsers(dest="cmd", required=True)
    adversarial.set_parser(subparsers)
    avg_model.set_parser(subparsers)
    convert_model.set_parser(subparsers)
    det_results.set_parser(subparsers)
    ensemble_model.set_parser(subparsers)
    download_model.set_parser(subparsers)
    introspection.set_parser(subparsers)
    labelme_to_coco.set_parser(subparsers)
    list_models.set_parser(subparsers)
    model_info.set_parser(subparsers)
    pack.set_parser(subparsers)
    quantize_model.set_parser(subparsers)
    results.set_parser(subparsers)
    show_det_iterator.set_parser(subparsers)
    show_iterator.set_parser(subparsers)
    similarity.set_parser(subparsers)
    stats.set_parser(subparsers)
    verify_coco.set_parser(subparsers)
    verify_directory.set_parser(subparsers)
    voc_to_coco.set_parser(subparsers)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



"""
VAN (Visual Attention Network), adapted from
https://github.com/Visual-Attention-Network/VAN-Classification/blob/main/models/van.py

Paper "Visual Attention Network", https://arxiv.org/abs/2202.09741
"""

# Reference license: Apache-2.0

import math
from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import StochasticDepth

from birder.layers import LayerNorm2d
from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class DWConvMLP(nn.Module):
    def __init__(self, in_features: int, hidden_features: int, drop: float) -> None:
        super().__init__()
        self.fc1 = nn.Conv2d(in_features, hidden_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.dwconv = nn.Conv2d(
            hidden_features,
            hidden_features,
            kernel_size=(3, 3),
            stride=(1, 1),
            padding=(1, 1),
            groups=hidden_features,
        )
        self.act = nn.GELU()
        self.fc2 = nn.Conv2d(hidden_features, in_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.drop = nn.Dropout(drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.fc1(x)
        x = self.dwconv(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class LKA(nn.Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.conv0 = nn.Conv2d(dim, dim, kernel_size=(5, 5), stride=(1, 1), padding=(2, 2), groups=dim)
        self.conv_spatial = nn.Conv2d(
            dim,
            dim,
            kernel_size=(7, 7),
            stride=(1, 1),
            padding=(9, 9),
            dilation=(3, 3),
            groups=dim,
        )
        self.conv1 = nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        u = x
        attn = self.conv0(x)
        attn = self.conv_spatial(attn)
        attn = self.conv1(attn)

        return u * attn


class Attention(nn.Module):
    def __init__(self, d_model: int) -> None:
        super().__init__()
        self.proj_1 = nn.Conv2d(d_model, d_model, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.activation = nn.GELU()
        self.spatial_gating_unit = LKA(d_model)
        self.proj_2 = nn.Conv2d(d_model, d_model, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        shorcut = x
        x = self.proj_1(x)
        x = self.activation(x)
        x = self.spatial_gating_unit(x)
        x = self.proj_2(x)
        x = x + shorcut

        return x


class OverlapPatchEmbed(nn.Module):
    def __init__(self, patch_size: tuple[int, int], stride: tuple[int, int], in_dim: int, embed_dim: int) -> None:
        super().__init__()
        self.proj = nn.Conv2d(
            in_dim, embed_dim, kernel_size=patch_size, stride=stride, padding=(patch_size[0] // 2, patch_size[1] // 2)
        )
        self.norm = nn.BatchNorm2d(embed_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.proj(x)
        x = self.norm(x)

        return x


class VANBlock(nn.Module):
    def __init__(self, dim: int, mlp_ratio: float, drop: float, drop_path: float) -> None:
        super().__init__()
        self.norm1 = nn.BatchNorm2d(dim)
        self.attn = Attention(dim)
        self.drop_path = StochasticDepth(drop_path, mode="row")

        self.norm2 = nn.BatchNorm2d(dim)
        self.mlp = DWConvMLP(in_features=dim, hidden_features=int(dim * mlp_ratio), drop=drop)

        layer_scale_init_value = 1e-2
        self.layer_scale_1 = nn.Parameter(layer_scale_init_value * torch.ones((1, dim, 1, 1)), requires_grad=True)
        self.layer_scale_2 = nn.Parameter(layer_scale_init_value * torch.ones((1, dim, 1, 1)), requires_grad=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.drop_path(self.layer_scale_1 * self.attn(self.norm1(x)))
        x = x + self.drop_path(self.layer_scale_2 * self.mlp(self.norm2(x)))

        return x


class VANStage(nn.Module):
    def __init__(
        self,
        patch_size: tuple[int, int],
        stride: tuple[int, int],
        in_dim: int,
        embed_dim: int,
        depth: int,
        mlp_ratio: float,
        drop: float,
        drop_path: list[float],
    ) -> None:
        super().__init__()
        self.patch_embed = OverlapPatchEmbed(patch_size=patch_size, stride=stride, in_dim=in_dim, embed_dim=embed_dim)

        layers = []
        for i in range(depth):
            layers.append(VANBlock(embed_dim, mlp_ratio=mlp_ratio, drop=drop, drop_path=drop_path[i]))

        self.block = nn.Sequential(*layers)
        self.norm = LayerNorm2d(embed_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.patch_embed(x)
        x = self.block(x)
        x = self.norm(x)

        return x


class VAN(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        drop_rate = 0.0
        mlp_ratios = [8.0, 8.0, 4.0, 4.0]
        embed_dims: list[int] = self.config["embed_dims"]
        depths: list[int] = self.config["depths"]
        drop_path_rate: float = self.config["drop_path_rate"]

        num_stages = len(depths)
        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths)).split(depths)]

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(num_stages):
            stages[f"stage{i+1}"] = VANStage(
                patch_size=(7, 7) if i == 0 else (3, 3),
                stride=(4, 4) if i == 0 else (2, 2),
                in_dim=self.input_channels if i == 0 else embed_dims[i - 1],
                embed_dim=embed_dims[i],
                depth=depths[i],
                mlp_ratio=mlp_ratios[i],
                drop=drop_rate,
                drop_path=dpr[i],
            )
            return_channels.append(embed_dims[i])

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = embed_dims[-1]
        self.classifier = self.create_classifier()

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

            elif isinstance(m, nn.LayerNorm):
                nn.init.zeros_(m.bias)
                nn.init.constant_(m.weight, 1.0)

            elif isinstance(m, nn.Conv2d):
                fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                fan_out //= m.groups
                nn.init.trunc_normal_(m.weight, std=math.sqrt(2.0 / fan_out))
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "van_b0", VAN, config={"embed_dims": [32, 64, 160, 256], "depths": [3, 3, 5, 2], "drop_path_rate": 0.1}
)
registry.register_model_config(
    "van_b1", VAN, config={"embed_dims": [64, 128, 320, 512], "depths": [2, 2, 4, 2], "drop_path_rate": 0.1}
)
registry.register_model_config(
    "van_b2", VAN, config={"embed_dims": [64, 128, 320, 512], "depths": [3, 3, 12, 3], "drop_path_rate": 0.1}
)
registry.register_model_config(
    "van_b3", VAN, config={"embed_dims": [64, 128, 320, 512], "depths": [3, 5, 27, 3], "drop_path_rate": 0.2}
)
registry.register_model_config(
    "van_b4", VAN, config={"embed_dims": [64, 128, 320, 512], "depths": [3, 6, 40, 3], "drop_path_rate": 0.3}
)
registry.register_model_config(
    "van_b5", VAN, config={"embed_dims": [96, 192, 480, 768], "depths": [3, 3, 24, 3], "drop_path_rate": 0.3}
)
registry.register_model_config(
    "van_b6", VAN, config={"embed_dims": [96, 192, 384, 768], "depths": [6, 6, 90, 6], "drop_path_rate": 0.5}
)

registry.register_weights(
    "van_b0_il-common",
    {
        "description": "VAN B0 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 15.2,
                "sha256": "6ab691627500cbba1e289e5293039630f76fba06b93222dddb5b06e711ccc31e",
            },
        },
        "net": {"network": "van_b0", "tag": "il-common"},
    },
)
registry.register_weights(
    "van_b2_arabian-peninsula256px",
    {
        "url": "https://huggingface.co/birder-project/van_b2_arabian-peninsula/resolve/main",
        "description": "VAN B2 model trained on the arabian-peninsula dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 101.2,
                "sha256": "fda34cdc4f87833dcb3eab29480305ea63d23575f0947c2757edbafed3ab5759",
            },
        },
        "net": {"network": "van_b2", "tag": "arabian-peninsula256px"},
    },
)
registry.register_weights(
    "van_b2_arabian-peninsula",
    {
        "url": "https://huggingface.co/birder-project/van_b2_arabian-peninsula/resolve/main",
        "description": "VAN B2 model trained on the arabian-peninsula dataset",
        "resolution": (384, 384),
        "formats": {
            "pt": {
                "file_size": 101.2,
                "sha256": "a119b788874a2dcf6454fb1357d7e128db450778fd2947d8b2bbdde5673a6cc7",
            },
        },
        "net": {"network": "van_b2", "tag": "arabian-peninsula"},
    },
)



"""
Paper "Very Deep Convolutional Networks for Large-Scale Image Recognition", https://arxiv.org/abs/1409.1556
"""

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
from torch import nn

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class Vgg(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        filters = [64, 128, 256, 512, 512]
        repeats: list[int] = self.config["repeats"]

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i, num in enumerate(repeats):
            layers = []
            for j in range(num):
                if i == 0 and j == 0:
                    in_channels = self.input_channels
                elif j == 0:
                    in_channels = filters[i - 1]
                else:
                    in_channels = filters[i]

                layers.append(
                    nn.Conv2d(
                        in_channels,
                        filters[i],
                        kernel_size=(3, 3),
                        stride=(1, 1),
                        padding=(1, 1),
                        bias=True,
                    )
                )
                layers.append(nn.ReLU(inplace=True))

            layers.append(nn.MaxPool2d(kernel_size=(2, 2), stride=(2, 2), padding=(0, 0)))
            stages[f"stage{i}"] = nn.Sequential(*layers)
            return_channels.append(filters[i])

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(7, 7)),
            nn.Flatten(1),
            nn.Linear(filters[-1] * 7 * 7, 4096),
            nn.ReLU(inplace=True),
            nn.Dropout(p=0.5),
            nn.Linear(4096, 4096),
            nn.ReLU(inplace=True),
            nn.Dropout(p=0.5),
        )
        self.return_channels = return_channels[1:5]
        self.embedding_size = 4096
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config("vgg_11", Vgg, config={"repeats": [1, 1, 2, 2, 2]})
registry.register_model_config("vgg_13", Vgg, config={"repeats": [2, 2, 2, 2, 2]})
registry.register_model_config("vgg_16", Vgg, config={"repeats": [2, 2, 3, 3, 3]})
registry.register_model_config("vgg_19", Vgg, config={"repeats": [2, 2, 4, 4, 4]})



"""
ViTDet backbone / ViT SAM, adapted from
https://github.com/facebookresearch/segment-anything/blob/main/segment_anything/modeling/image_encoder.py

Paper "Exploring Plain Vision Transformer Backbones for Object Detection",
https://arxiv.org/abs/2203.16527

and as used as an image encoder at the paper "Segment Anything", https://arxiv.org/abs/2304.02643
"""

# Reference license: Apache-2.0

from collections.abc import Callable
from functools import partial
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import StochasticDepth

from birder.layers import FFN
from birder.layers import LayerNorm2d
from birder.layers import LayerScale
from birder.layers import SwiGLU_FFN
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.vit import EncoderBlock as MAEDecoderBlock


# pylint: disable=invalid-name
def window_partition(x: torch.Tensor, window_size: int) -> tuple[torch.Tensor, tuple[int, int]]:
    (B, H, W, C) = x.shape

    pad_h = (window_size - H % window_size) % window_size
    pad_w = (window_size - W % window_size) % window_size
    if pad_h > 0 or pad_w > 0:
        x = F.pad(x, (0, 0, 0, pad_w, 0, pad_h))

    Hp = H + pad_h
    Wp = W + pad_w

    x = x.view(B, Hp // window_size, window_size, Wp // window_size, window_size, C)
    windows = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(-1, window_size, window_size, C)

    return (windows, (Hp, Wp))


# pylint: disable=invalid-name
def window_unpartition(
    windows: torch.Tensor, window_size: int, pad_hw: tuple[int, int], hw: tuple[int, int]
) -> torch.Tensor:
    (Hp, Wp) = pad_hw
    (H, W) = hw
    B = windows.shape[0] // (Hp * Wp // window_size // window_size)
    x = windows.view(B, Hp // window_size, Wp // window_size, window_size, window_size, -1)
    x = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(B, Hp, Wp, -1)

    if Hp > H or Wp > W:
        x = x[:, :H, :W, :].contiguous()

    return x


def get_rel_pos(q_size: int, k_size: int, rel_pos: torch.Tensor) -> torch.Tensor:
    max_rel_dist = int(2 * max(q_size, k_size) - 1)

    # Interpolate rel pos if needed
    if rel_pos.shape[0] != max_rel_dist:
        # Adjust size is a one off interpolation, should prevent us from getting here
        rel_pos_resized = F.interpolate(
            rel_pos.reshape(1, rel_pos.shape[0], -1).permute(0, 2, 1), size=max_rel_dist, mode="linear"
        )
        rel_pos_resized = rel_pos_resized.reshape(-1, max_rel_dist).permute(1, 0)
    else:
        rel_pos_resized = rel_pos

    # Scale the coords with short length if shapes for q and k are different
    q_coords = torch.arange(q_size)[:, None] * max(k_size / q_size, 1.0)
    k_coords = torch.arange(k_size)[None, :] * max(q_size / k_size, 1.0)
    relative_coords = (q_coords - k_coords) + (k_size - 1) * max(q_size / k_size, 1.0)

    return rel_pos_resized[relative_coords.long()]


def get_decomposed_rel_pos_bias(
    q: torch.Tensor, rel_pos_h: torch.Tensor, rel_pos_w: torch.Tensor, q_size: tuple[int, int], k_size: tuple[int, int]
) -> torch.Tensor:
    (q_h, q_w) = q_size
    (k_h, k_w) = k_size
    Rh = get_rel_pos(q_h, k_h, rel_pos_h)
    Rw = get_rel_pos(q_w, k_w, rel_pos_w)

    (B, _, dim) = q.shape
    r_q = q.reshape(B, q_h, q_w, dim)
    rel_h = torch.einsum("bhwc,hkc->bhwk", r_q, Rh)
    rel_w = torch.einsum("bhwc,wkc->bhwk", r_q, Rw)

    attn_bias = rel_h[:, :, :, :, None] + rel_w[:, :, :, None, :]
    attn_bias = attn_bias.reshape(-1, q_h * q_w, k_h * k_w)

    return attn_bias


class PatchEmbed(nn.Module):
    def __init__(self, in_channels: int, embed_dim: int, kernel_size: tuple[int, int], stride: tuple[int, int]) -> None:
        super().__init__()
        self.proj = nn.Conv2d(in_channels, embed_dim, kernel_size=kernel_size, stride=stride, padding=(0, 0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.proj(x)

        # B C H W -> B H W C
        return x.permute(0, 2, 3, 1)


class Attention(nn.Module):
    def __init__(
        self, dim: int, num_heads: int, qkv_bias: bool, use_rel_pos: bool, input_size: Optional[tuple[int, int]] = None
    ) -> None:
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.proj = nn.Linear(dim, dim)

        self.use_rel_pos = use_rel_pos
        if self.use_rel_pos is True:
            assert input_size is not None, "Input size must be provided if using relative positional encoding."
            self.rel_pos_h = nn.Parameter(torch.zeros(2 * input_size[0] - 1, head_dim))
            self.rel_pos_w = nn.Parameter(torch.zeros(2 * input_size[1] - 1, head_dim))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, H, W, _) = x.shape
        qkv = self.qkv(x).reshape(B, H * W, 3, self.num_heads, -1).permute(2, 0, 3, 1, 4)
        (q, k, v) = qkv.reshape(3, B * self.num_heads, H * W, -1).unbind(0)

        if self.use_rel_pos is True:
            attn_bias = get_decomposed_rel_pos_bias(q, self.rel_pos_h, self.rel_pos_w, (H, W), (H, W))
        else:
            attn_bias = None

        x = F.scaled_dot_product_attention(  # pylint: disable=not-callable
            q, k, v, attn_mask=attn_bias, scale=self.scale
        )

        x = x.view(B, self.num_heads, H, W, -1).permute(0, 2, 3, 1, 4).reshape(B, H, W, -1)
        x = self.proj(x)

        return x


class EncoderBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        mlp_ratio: float,
        qkv_bias: bool,
        drop_path: float,
        use_rel_pos: bool,
        window_size: int,
        input_size: Optional[tuple[int, int]] = None,
        activation_layer: Callable[..., nn.Module] = nn.GELU,
        layer_scale_init_value: Optional[float] = None,
        norm_layer: Callable[..., nn.Module] = nn.LayerNorm,
        mlp_layer: Callable[..., nn.Module] = FFN,
    ) -> None:
        super().__init__()
        self.window_size = window_size

        # Attention block
        self.norm1 = norm_layer(dim, eps=1e-6)
        self.attn = Attention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            use_rel_pos=use_rel_pos,
            input_size=input_size if window_size == 0 else (window_size, window_size),
        )
        self.drop_path1 = StochasticDepth(drop_path, mode="row")
        if layer_scale_init_value is not None:
            self.layer_scale_1 = LayerScale(dim, layer_scale_init_value)
        else:
            self.layer_scale_1 = nn.Identity()

        # MLP block
        self.norm2 = norm_layer(dim, eps=1e-6)
        self.mlp = mlp_layer(dim, int(dim * mlp_ratio), act_layer=activation_layer, dropout=0.0)
        self.drop_path2 = StochasticDepth(drop_path, mode="row")
        if layer_scale_init_value is not None:
            self.layer_scale_2 = LayerScale(dim, layer_scale_init_value)
        else:
            self.layer_scale_2 = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (_, H, W, _) = x.shape
        shortcut = x

        x = self.norm1(x)
        pad_hw = (0, 0)
        if self.window_size > 0:
            (x, pad_hw) = window_partition(x, self.window_size)

        x = self.attn(x)
        if self.window_size > 0:
            x = window_unpartition(x, self.window_size, pad_hw, (H, W))

        x = self.layer_scale_1(x)
        x = self.drop_path1(x) + shortcut
        x = x + self.drop_path2(self.layer_scale_2(self.mlp(self.norm2(x))))

        return x


# pylint: disable=invalid-name
class ViT_SAM(DetectorBackbone):
    block_group_regex = r"body\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        image_size = self.size
        use_rel_pos = True
        patch_size: int = self.config["patch_size"]
        num_layers: int = self.config["num_layers"]
        num_heads: int = self.config["num_heads"]
        hidden_dim: int = self.config["hidden_dim"]
        mlp_ratio: float = self.config["mlp_ratio"]
        layer_scale_init_value: Optional[float] = self.config.get("layer_scale_init_value", None)
        norm_layer_type: str = self.config.get("norm_layer_type", "LayerNorm")
        mlp_layer_type: str = self.config.get("mlp_layer_type", "FFN")
        window_size: int = self.config["window_size"]
        global_attn_indexes: list[int] = self.config["global_attn_indexes"]
        neck_channels: Optional[int] = self.config.get("neck_channels", None)
        drop_path_rate: float = self.config["drop_path_rate"]

        if norm_layer_type == "LayerNorm":
            norm_layer = nn.LayerNorm
        elif norm_layer_type == "RMSNorm":
            norm_layer = nn.RMSNorm
        else:
            raise ValueError(f"Unknown norm_layer_type '{norm_layer_type}'")

        if mlp_layer_type == "FFN":
            mlp_layer = FFN
            act_layer = nn.GELU
        elif mlp_layer_type == "SwiGLU_FFN":
            mlp_layer = SwiGLU_FFN
            act_layer = nn.SiLU
        else:
            raise ValueError(f"Unknown mlp_layer_type '{mlp_layer_type}'")

        torch._assert(image_size[0] % patch_size == 0, "Input shape indivisible by patch size!")
        torch._assert(image_size[1] % patch_size == 0, "Input shape indivisible by patch size!")
        torch._assert(hidden_dim % num_heads == 0, "Hidden dim indivisible by num heads!")
        self.patch_size = patch_size
        self.num_layers = num_layers
        self.hidden_dim = hidden_dim
        self.global_attn_indexes = global_attn_indexes
        self.num_special_tokens = 0
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, num_layers)]  # Stochastic depth decay rule

        self.patch_embed = PatchEmbed(
            in_channels=self.input_channels,
            embed_dim=hidden_dim,
            kernel_size=(patch_size, patch_size),
            stride=(patch_size, patch_size),
        )

        # Absolute position embedding
        self.pos_embedding = nn.Parameter(
            torch.zeros(1, image_size[0] // patch_size, image_size[1] // patch_size, hidden_dim)
        )

        layers = []
        for i in range(num_layers):
            layers.append(
                EncoderBlock(
                    dim=hidden_dim,
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=True,
                    drop_path=dpr[i],
                    use_rel_pos=use_rel_pos,
                    window_size=window_size if i not in global_attn_indexes else 0,
                    input_size=(image_size[0] // patch_size, image_size[1] // patch_size),
                    activation_layer=act_layer,
                    layer_scale_init_value=layer_scale_init_value,
                    norm_layer=norm_layer,
                    mlp_layer=mlp_layer,
                )
            )

        self.body = nn.Sequential(*layers)
        if neck_channels is not None:
            self.neck = nn.Sequential(
                nn.Conv2d(
                    hidden_dim,
                    neck_channels,
                    kernel_size=(1, 1),
                    stride=(1, 1),
                    padding=(0, 0),
                    bias=False,
                ),
                LayerNorm2d(neck_channels),
                nn.Conv2d(
                    neck_channels,
                    neck_channels,
                    kernel_size=(3, 3),
                    stride=(1, 1),
                    padding=(1, 1),
                    bias=False,
                ),
                LayerNorm2d(neck_channels),
            )
        else:
            neck_channels = hidden_dim
            self.neck = LayerNorm2d(neck_channels)

        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )

        self.return_stages = ["neck"]  # Actually meaningless, but for completeness
        self.return_channels = [neck_channels]
        self.embedding_size = neck_channels
        self.classifier = self.create_classifier()

        self.encoding_size = hidden_dim * (image_size[0] // patch_size) * (image_size[1] // patch_size)
        self.decoder_block = partial(
            MAEDecoderBlock,
            16,
            mlp_dim=None,
            dropout=0,
            attention_dropout=0,
            drop_path=0,
            activation_layer=nn.GELU,
        )

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.patch_embed(x)
        x = x + self.pos_embedding

        x = self.body(x)
        x = self.neck(x.permute(0, 3, 1, 2))
        return {self.return_stages[0]: x}

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.patch_embed.parameters():
            param.requires_grad = False

        self.pos_embedding.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.patch_embed(x)
        x = x + self.pos_embedding

        x = self.body(x)
        x = self.neck(x.permute(0, 3, 1, 2))

        return x

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        assert dynamic_size is False, "Dynamic size not supported for this network"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        assert new_size[0] % self.patch_size == 0, "Input shape indivisible by patch size!"
        assert new_size[1] % self.patch_size == 0, "Input shape indivisible by patch size!"

        super().adjust_size(new_size)

        new_base_size_h = new_size[0] // self.patch_size
        new_base_size_w = new_size[1] // self.patch_size
        for idx, m in enumerate(self.body):
            if idx not in self.global_attn_indexes:
                continue

            if m.attn.use_rel_pos is True:
                rel_pos_h = m.attn.rel_pos_h
                rel_pos_w = m.attn.rel_pos_w

                orig_dtype = rel_pos_h.dtype
                rel_pos_h = rel_pos_h.float()
                rel_pos_w = rel_pos_w.float()

                rel_pos_h = rel_pos_h.permute(1, 0)
                rel_pos_w = rel_pos_w.permute(1, 0)
                rel_pos_h = rel_pos_h.unsqueeze(0)
                rel_pos_w = rel_pos_w.unsqueeze(0)

                rel_pos_h = F.interpolate(rel_pos_h, size=2 * new_base_size_h - 1, mode="linear")
                rel_pos_w = F.interpolate(rel_pos_w, size=2 * new_base_size_w - 1, mode="linear")
                rel_pos_h = rel_pos_h.squeeze(0)
                rel_pos_w = rel_pos_w.squeeze(0)
                rel_pos_h = rel_pos_h.permute(1, 0)
                rel_pos_w = rel_pos_w.permute(1, 0)
                rel_pos_h = rel_pos_h.to(orig_dtype)
                rel_pos_w = rel_pos_w.to(orig_dtype)

                m.attn.rel_pos_h = nn.Parameter(rel_pos_h)
                m.attn.rel_pos_w = nn.Parameter(rel_pos_w)

        orig_dtype = self.pos_embedding.dtype
        pos_embedding = self.pos_embedding.float()
        pos_embedding = pos_embedding.permute(0, 3, 1, 2)
        pos_embedding = F.interpolate(
            pos_embedding, size=(new_base_size_h, new_base_size_w), mode="bicubic", antialias=True
        )
        pos_embedding = pos_embedding.permute(0, 2, 3, 1)
        pos_embedding = pos_embedding.to(orig_dtype)

        self.pos_embedding = nn.Parameter(pos_embedding)

    def load_vit_weights(self, state_dict: dict[str, Any]) -> None:
        """
        As suggested at "Exploring Plain Vision Transformer Backbones for Object Detection"
        (https://arxiv.org/abs/2203.16527), weights can be transfer from vanilla ViT for faster start-up.
        The relative position embedding and "neck" convolutions are left intact.
        """

        # Remove all special token
        num_special_tokens = 0
        if "class_token" in state_dict:
            num_special_tokens += 1
            del state_dict["class_token"]

        if "reg_tokens" in state_dict:
            num_special_tokens += state_dict["reg_tokens"].size(1)
            del state_dict["reg_tokens"]

        # Remove final norm
        del state_dict["norm.weight"]
        if "norm.bias" in state_dict:
            del state_dict["norm.bias"]

        # Remove classifier weights
        if "classifier.weight" in state_dict:
            del state_dict["classifier.weight"]
            del state_dict["classifier.bias"]

        seq_length = (self.size[0] // self.patch_size) * (self.size[1] // self.patch_size)

        # Adjust pos_embedding
        if state_dict["pos_embedding"].ndim == 2:
            vit_pos_embed_special_tokens = state_dict["pos_embedding"].size(0) != seq_length
            if vit_pos_embed_special_tokens is True:
                state_dict["pos_embedding"] = state_dict["pos_embedding"][num_special_tokens:, :]
        else:
            vit_pos_embed_special_tokens = state_dict["pos_embedding"].size(1) != seq_length
            if vit_pos_embed_special_tokens is True:
                state_dict["pos_embedding"] = state_dict["pos_embedding"][:, num_special_tokens:, :]

        state_dict["pos_embedding"] = state_dict["pos_embedding"].reshape(
            1, self.size[0] // self.patch_size, self.size[1] // self.patch_size, -1
        )

        # Modify encoder weight names
        modified_state_dict = {}
        for k, v in state_dict.items():
            if k.startswith("conv_proj.") is True:
                k = k.replace("conv_proj.", "patch_embed.proj.")
            elif k.startswith("encoder.block.") is True:
                k = k.replace("encoder.block.", "body.")
                k = k.replace("ln1", "norm1")
                k = k.replace("ln2", "norm2")
                if "self_attention.in_proj" in k:
                    k = k.replace("self_attention.in_proj_", "attn.qkv.")
                elif "self_attention.out_proj" in k:
                    k = k.replace("self_attention.out_proj.", "attn.proj.")

            modified_state_dict[k] = v

        # Load the modified state dict
        incompatible_keys = self.load_state_dict(modified_state_dict, strict=False)
        assert len(incompatible_keys.unexpected_keys) == 0


# ViTDet (no neck)
registry.register_model_config(
    "vit_det_m16_rms",
    ViT_SAM,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 8,
        "hidden_dim": 512,
        "mlp_ratio": 4.0,
        "norm_layer_type": "RMSNorm",
        "window_size": 14,
        "global_attn_indexes": [2, 5, 8, 11],
        "drop_path_rate": 0.0,
    },
)

registry.register_model_config(
    "vit_det_b16",
    ViT_SAM,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_ratio": 4.0,
        "window_size": 14,
        "global_attn_indexes": [2, 5, 8, 11],
        "drop_path_rate": 0.1,
    },
)

# ViT SAM (with neck)
registry.register_model_config(
    "vit_sam_b16",
    ViT_SAM,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_ratio": 4.0,
        "window_size": 14,
        "global_attn_indexes": [2, 5, 8, 11],
        "neck_channels": 256,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_sam_l16",
    ViT_SAM,
    config={
        "patch_size": 16,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_ratio": 4.0,
        "window_size": 14,
        "global_attn_indexes": [5, 11, 17, 23],
        "neck_channels": 256,
        "drop_path_rate": 0.4,
    },
)
registry.register_model_config(
    "vit_sam_h16",
    ViT_SAM,
    config={
        "patch_size": 16,
        "num_layers": 32,
        "num_heads": 16,
        "hidden_dim": 1280,
        "mlp_ratio": 4.0,
        "window_size": 14,
        "global_attn_indexes": [7, 15, 23, 31],
        "neck_channels": 256,
        "drop_path_rate": 0.5,
    },
)



import argparse
import json
import logging
import time
from typing import Any

import torch
import torch.ao.quantization
from torch.ao.quantization.quantize_fx import convert_fx
from torch.ao.quantization.quantize_fx import prepare_fx
from torch.utils.data import DataLoader
from torch.utils.data import Subset
from torch.utils.mobile_optimizer import optimize_for_mobile
from torchvision.datasets import ImageFolder
from tqdm import tqdm

from birder.common import cli
from birder.common import fs_ops
from birder.common import lib
from birder.common.lib import get_network_name
from birder.conf import settings
from birder.data.transforms.classification import inference_preset

logger = logging.getLogger(__name__)


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "quantize-model",
        allow_abbrev=False,
        help="quantize model",
        description="quantize model",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools quantize-model --network shufflenet_v2_2_0 --epoch 0\n"
            "python -m birder.tools quantize-model -n convnext_v2_tiny -e 0 --qbackend x86 \n"
            "python -m birder.tools quantize-model --network densenet_121 -e 100 --num-calibration-batches 256\n"
            "python -m birder.tools quantize-model -n efficientnet_v2_s -e 200 --qbackend x86\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument(
        "-n", "--network", type=str, required=True, help="the neural network to load (i.e. resnet_v2_50)"
    )
    subparser.add_argument("-e", "--epoch", type=int, metavar="N", help="model checkpoint to load")
    subparser.add_argument("-t", "--tag", type=str, help="model tag (from the training phase)")
    subparser.add_argument(
        "-r", "--reparameterized", default=False, action="store_true", help="load reparameterized model"
    )
    subparser.add_argument("-f", "--force", action="store_true", help="override existing model")
    subparser.add_argument(
        "-j",
        "--num-workers",
        type=int,
        default=8,
        help="number of preprocessing workers",
    )
    subparser.add_argument(
        "--qbackend",
        type=str,
        choices=["qnnpack", "x86", "onednn"],
        default="qnnpack",
        help="quantized backend",
    )
    subparser.add_argument("--batch-size", type=int, default=32, metavar="N", help="the batch size")
    subparser.add_argument(
        "--num-calibration-batches",
        default=128,
        type=int,
        help="number of batches of training set for observer calibration",
    )
    subparser.add_argument("--gpu", default=False, action="store_true", help="use gpu")
    subparser.add_argument(
        "--data-path", type=str, default=str(settings.TRAINING_DATA_PATH), help="training directory path"
    )
    subparser.set_defaults(func=main)


def main(args: argparse.Namespace) -> None:
    network_name = get_network_name(args.network, tag=args.tag)
    model_path = fs_ops.model_path(network_name, epoch=args.epoch, quantized=True, pts=True)
    if model_path.exists() is True and args.force is False:
        logger.warning("Quantized model already exists... aborting")
        raise SystemExit(1)

    if args.gpu is True:
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")

    # Load model
    (net, (class_to_idx, signature, rgb_stats, *_)) = fs_ops.load_model(
        device, args.network, tag=args.tag, epoch=args.epoch, inference=True, reparameterized=args.reparameterized
    )
    net.eval()
    task = net.task
    size = lib.get_size_from_signature(signature)

    # Set calibration data
    full_dataset = ImageFolder(args.data_path, transform=inference_preset(size, rgb_stats, 1.0))
    calibration_dataset = Subset(full_dataset, indices=list(range(args.batch_size * args.num_calibration_batches)))
    calibration_data_loader = DataLoader(
        calibration_dataset,
        batch_size=args.batch_size,
        shuffle=False,
        num_workers=args.num_workers,
        pin_memory=True,
    )

    # Quantization
    tic = time.time()
    qconfig_mapping = torch.ao.quantization.get_default_qconfig_mapping(args.qbackend)

    example_inputs = next(iter(calibration_data_loader))[0]
    prepared_net = prepare_fx(net, qconfig_mapping, example_inputs)

    with tqdm(total=len(calibration_dataset), initial=0, unit="images", unit_scale=True, leave=False) as progress:
        with torch.inference_mode():
            for inputs, targets in calibration_data_loader:
                inputs = inputs.to(device)
                targets = targets.to(device)
                _ = prepared_net(inputs)

                # Update progress bar
                progress.update(n=args.batch_size)

    net = convert_fx(prepared_net)

    toc = time.time()
    (minutes, seconds) = divmod(toc - tic, 60)
    logger.info(f"{int(minutes):0>2}m{seconds:04.1f}s to quantize model")

    model_path = fs_ops.model_path(network_name, epoch=args.epoch, quantized=True, pts=True)
    logger.info(f"Saving quantized TorchScript model {model_path}...")

    # Convert to TorchScript
    scripted_module = torch.jit.script(net)
    fs_ops.save_pts(scripted_module, model_path, task, class_to_idx, signature, rgb_stats)

    if args.qbackend == "qnnpack":
        model_path = fs_ops.model_path(network_name, epoch=args.epoch, quantized=True, lite=True)
        if model_path.exists() is True and args.force is False:
            logger.warning("Quantized model lite already exists... aborting")
            raise SystemExit(1)

        logger.info(f"Saving quantized TorchScript model {model_path}...")
        optimized_scripted_module = optimize_for_mobile(scripted_module)
        optimized_scripted_module._save_for_lite_interpreter(  # pylint: disable=protected-access
            str(model_path),
            _extra_files={
                "task": task,
                "class_to_idx": json.dumps(class_to_idx),
                "signature": json.dumps(signature),
                "rgb_stats": json.dumps(rgb_stats),
            },
        )



import os
import random
from typing import Any
from typing import Optional

import numpy as np
import torch

from birder.conf import settings
from birder.data.transforms.classification import RGBType
from birder.model_registry import registry
from birder.net.base import BaseNet
from birder.net.base import SignatureType
from birder.net.base import reparameterize_available
from birder.net.detection.base import DetectionBaseNet
from birder.net.detection.base import DetectionSignatureType
from birder.net.mim.base import MIMBaseNet
from birder.net.ssl.base import SSLBaseNet
from birder.version import __version__


def set_random_seeds(seed: int) -> None:
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)


def get_size_from_signature(signature: SignatureType | DetectionSignatureType) -> tuple[int, int]:
    return tuple(signature["inputs"][0]["data_shape"][2:4])  # type: ignore[return-value]


def get_channels_from_signature(signature: SignatureType | DetectionSignatureType) -> int:
    return signature["inputs"][0]["data_shape"][1]


def get_num_labels_from_signature(signature: SignatureType | DetectionSignatureType) -> int:
    if "num_labels" in signature:
        return signature["num_labels"]  # type: ignore[typeddict-item]

    return signature["outputs"][0]["data_shape"][1]


def get_label_from_path(path: str, hierarchical: bool = False, root: str = "", separator: str = "_") -> str:
    """
    Returns the last directory from the path.

    Standard mode (hierarchical=False):
        data/validation/Alpine swift/000001.jpeg -> 'Alpine swift'
        data/train/cats/img1.jpg -> 'cats'

    Hierarchical mode (hierarchical=True):
        data/validation/Aves/Alpine swift/000001.jpeg -> 'Aves_Alpine swift'
        data/train/animals/cats/img1.jpg -> 'animals_cats'
    """

    if hierarchical is False:
        return os.path.basename(os.path.dirname(path))

    directory = os.path.dirname(path)
    rel_path = os.path.relpath(directory, root)
    return rel_path.replace(os.sep, separator)


def get_network_name(network: str, tag: Optional[str] = None) -> str:
    network_name = network
    if tag is not None:
        network_name = f"{network_name}_{tag}"

    return network_name


def get_mim_network_name(network: str, encoder: str, tag: Optional[str] = None) -> str:
    prefix = get_network_name(network)
    suffix = get_network_name(encoder, tag)
    return f"{prefix}_{suffix}"


def get_detection_network_name(network: str, tag: Optional[str], backbone: str, backbone_tag: Optional[str]) -> str:
    prefix = get_network_name(network, tag)
    suffix = get_network_name(backbone, backbone_tag)
    return f"{prefix}_{suffix}"


def detection_class_to_idx(class_to_idx: dict[str, int]) -> dict[str, int]:
    # Give place to "background" class (always index 0)
    for key in class_to_idx:
        class_to_idx[key] += 1

    return class_to_idx


def class_to_idx_from_coco(cats: dict[int, Any]) -> dict[str, int]:
    class_list = [item["name"] for item in sorted(cats.values(), key=lambda x: x["id"])]
    class_to_idx = {k: v for v, k in enumerate(class_list)}
    return class_to_idx


def get_network_config(
    net: BaseNet | DetectionBaseNet | MIMBaseNet | SSLBaseNet,
    signature: SignatureType | DetectionSignatureType,
    rgb_stats: RGBType,
) -> dict[str, Any]:
    model_name = registry.get_model_base_name(net)
    alias = registry.get_model_alias(net)
    model_config = None
    if net.config is not None:
        model_config = net.config

    backbone_config: dict[str, Any] = {}
    if isinstance(net, DetectionBaseNet):
        backbone_config["backbone"] = registry.get_model_base_name(net.backbone)
        backbone_config["backbone_alias"] = registry.get_model_alias(net.backbone)
        if net.backbone.config is not None:
            backbone_config["backbone_config"] = net.backbone.config
        if reparameterize_available(net.backbone) is True:
            backbone_config["backbone_reparameterized"] = net.backbone.reparameterized

    encoder_config: dict[str, Any] = {}
    if isinstance(net, MIMBaseNet):
        encoder_config["encoder"] = registry.get_model_base_name(net.encoder)
        encoder_config["encoder_alias"] = registry.get_model_alias(net.encoder)
        if net.encoder.config is not None:
            encoder_config["encoder_config"] = net.encoder.config
        if reparameterize_available(net.encoder) is True:
            encoder_config["encoder_reparameterized"] = net.encoder.reparameterized

    base_net_config: dict[str, Any] = {}
    if isinstance(net, SSLBaseNet):
        base_net_config["base_net"] = registry.get_model_base_name(net.backbone)
        base_net_config["base_net_alias"] = registry.get_model_alias(net.backbone)
        if net.backbone.config is not None:
            base_net_config["base_net_config"] = net.backbone.config
        if reparameterize_available(net.backbone) is True:
            base_net_config["base_net_reparameterized"] = net.backbone.reparameterized

    net_config = {
        "birder_version": __version__,
        "name": model_name,
        "alias": alias,
        "task": net.task,
        "model_config": model_config,
        **backbone_config,
        **encoder_config,
        **base_net_config,
        "signature": signature,
        "rgb_stats": rgb_stats,
    }

    if reparameterize_available(net) is True:
        net_config["reparameterized"] = net.reparameterized

    return net_config


def get_pretrained_model_url(weights: str, file_format: str) -> tuple[str, str]:
    model_metadata = registry.get_pretrained_metadata(weights)
    model_file = f"{weights}.{file_format}"
    base_url = model_metadata.get("url", settings.REGISTRY_BASE_UTL)
    url = f"{base_url}/{model_file}"

    return (model_file, url)


def format_duration(seconds: float) -> str:
    s = int(seconds)
    (mm, ss) = divmod(s, 60)
    (hh, mm) = divmod(mm, 60)
    return f"{hh:d}:{mm:02d}:{ss:02d}"



"""
StarNet, adapted from
https://github.com/ma-xu/Rewrite-the-Stars/blob/main/imagenet/starnet.py

Paper "Rewrite the Stars", https://arxiv.org/abs/2403.19967

Changes from original:
* Removed bias term from conv before batchnorm
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import StochasticDepth

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class StarNetBlock(nn.Module):
    def __init__(self, dim: int, mlp_ratio: float, drop_path: float) -> None:
        super().__init__()
        self.dwconv1 = Conv2dNormActivation(
            dim, dim, kernel_size=(7, 7), stride=(1, 1), padding=(3, 3), groups=dim, activation_layer=None
        )
        self.f1 = nn.Conv2d(dim, int(mlp_ratio * dim), kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.f2 = nn.Conv2d(dim, int(mlp_ratio * dim), kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.g = Conv2dNormActivation(
            int(mlp_ratio * dim), dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), activation_layer=None
        )
        self.dwconv2 = nn.Conv2d(dim, dim, kernel_size=(7, 7), stride=(1, 1), padding=(3, 3), groups=dim)
        self.act = nn.ReLU6()
        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x
        x = self.dwconv1(x)
        x1 = self.f1(x)
        x2 = self.f2(x)
        x = self.act(x1) * x2
        x = self.dwconv2(self.g(x))
        x = identity + self.drop_path(x)

        return x


class StarNetStage(nn.Module):
    def __init__(self, in_channels: int, embed_dim: int, depth: int, mlp_ratio: float, drop_path: list[float]) -> None:
        super().__init__()

        self.downsample = Conv2dNormActivation(
            in_channels, embed_dim, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), activation_layer=None
        )
        layers = []
        for i in range(depth):
            layers.append(StarNetBlock(embed_dim, mlp_ratio=mlp_ratio, drop_path=drop_path[i]))

        self.blocks = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.blocks(x)

        return x


class StarNet(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        first_channels = 32
        base_dim: int = self.config["base_dim"]
        depths: list[int] = self.config["depths"]
        mlp_ratio: float = self.config["mlp_ratio"]
        drop_path_rate: float = self.config["drop_path_rate"]

        self.stem = Conv2dNormActivation(
            self.input_channels,
            first_channels,
            kernel_size=(3, 3),
            stride=(2, 2),
            padding=(1, 1),
            activation_layer=nn.ReLU6,
        )

        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths)).split(depths)]
        num_stages = len(depths)

        prev_channels = first_channels
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(num_stages):
            embed_dim = base_dim * 2**i
            stages[f"stage{i+1}"] = StarNetStage(
                prev_channels, embed_dim=embed_dim, depth=depths[i], mlp_ratio=mlp_ratio, drop_path=dpr[i]
            )
            return_channels.append(embed_dim)
            prev_channels = embed_dim

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.BatchNorm2d(embed_dim),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = embed_dim
        self.classifier = self.create_classifier()

        # Weight initialization
        for m in self.modules():
            if isinstance(m, (nn.Conv2d, nn.Linear)):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, nn.BatchNorm2d):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


# Extremely Small Models
registry.register_model_config(
    "starnet_esm05", StarNet, config={"base_dim": 16, "depths": [1, 1, 3, 1], "mlp_ratio": 3.0, "drop_path_rate": 0.0}
)
registry.register_model_config(
    "starnet_esm10", StarNet, config={"base_dim": 20, "depths": [1, 2, 4, 1], "mlp_ratio": 4.0, "drop_path_rate": 0.0}
)
registry.register_model_config(
    "starnet_esm15", StarNet, config={"base_dim": 24, "depths": [1, 2, 4, 2], "mlp_ratio": 3.0, "drop_path_rate": 0.0}
)

# StarNets
registry.register_model_config(
    "starnet_s1", StarNet, config={"base_dim": 24, "depths": [2, 2, 8, 3], "mlp_ratio": 4.0, "drop_path_rate": 0.0}
)
registry.register_model_config(
    "starnet_s2", StarNet, config={"base_dim": 32, "depths": [1, 2, 6, 2], "mlp_ratio": 4.0, "drop_path_rate": 0.0}
)
registry.register_model_config(
    "starnet_s3", StarNet, config={"base_dim": 32, "depths": [2, 2, 8, 4], "mlp_ratio": 4.0, "drop_path_rate": 0.0}
)
registry.register_model_config(
    "starnet_s4", StarNet, config={"base_dim": 32, "depths": [3, 3, 12, 5], "mlp_ratio": 4.0, "drop_path_rate": 0.1}
)

registry.register_weights(
    "starnet_esm10_il-common",
    {
        "description": "StarNet extremely small model 1M trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 3.7,
                "sha256": "74f49aff37993c8cb3d76adba72318a37e09816173505b180bb4353699a4dbdc",
            }
        },
        "net": {"network": "starnet_esm10", "tag": "il-common"},
    },
)
registry.register_weights(
    "starnet_s1_il-common",
    {
        "description": "StarNet S1 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 10.6,
                "sha256": "b997d7ec49c1b6dc1dbc1f12204d34d1fce8114072e7e72b9fe6df989863143b",
            }
        },
        "net": {"network": "starnet_s1", "tag": "il-common"},
    },
)



"""
Generalized Mean (GeM) Pooling Layers

These modules implement Generalized Mean Pooling as described in the paper
"Fine-tuning CNN Image Retrieval with No Human Annotation"
https://arxiv.org/abs/1711.02512

Two variants are provided:
1. FixedGeMPool2d: Performs global GeM pooling with a fixed, non-trainable pooling parameter p.
2. GeMPool2d: Performs global GeM pooling with a trainable pooling parameter "p".
   This is often used as the final pooling layer in image retrieval networks.
"""

import torch
import torch.nn.functional as F
from torch import nn


class FixedGeMPool2d(nn.Module):
    def __init__(self, pooling_param: float, eps: float = 1e-6) -> None:
        super().__init__()
        self.p = pooling_param
        self.eps = eps

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x.flatten(2)
        mean = x.clamp(min=self.eps).pow(self.p).mean(dim=2)

        return mean.pow(1.0 / self.p)


class GeMPool2d(nn.Module):
    def __init__(self, pooling_param: float, eps: float = 1e-6):
        super().__init__()
        self.p = nn.Parameter(torch.ones(1) * pooling_param)
        self.eps = eps

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        mean = F.avg_pool2d(x.clamp(min=self.eps).pow(self.p), (x.size(-2), x.size(-1)))  # pylint:disable=not-callable
        return mean.pow(1.0 / self.p).squeeze(dim=(-2, -1))



"""
Paper "Squeeze-and-Excitation Networks", https://arxiv.org/abs/1709.01507
"""

from typing import Any
from typing import Optional

from birder.model_registry import registry
from birder.net.resnet_v2 import ResNet_v2


# pylint: disable=invalid-name
class SE_ResNet_v2(ResNet_v2):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size, squeeze_excitation=True)


registry.register_model_config(
    "se_resnet_v2_18",
    SE_ResNet_v2,
    config={"bottle_neck": False, "filter_list": [64, 64, 128, 256, 512], "units": [2, 2, 2, 2]},
)
registry.register_model_config(
    "se_resnet_v2_34",
    SE_ResNet_v2,
    config={"bottle_neck": False, "filter_list": [64, 64, 128, 256, 512], "units": [3, 4, 6, 3]},
)
registry.register_model_config(
    "se_resnet_v2_50",
    SE_ResNet_v2,
    config={"bottle_neck": True, "filter_list": [64, 256, 512, 1024, 2048], "units": [3, 4, 6, 3]},
)
registry.register_model_config(
    "se_resnet_v2_101",
    SE_ResNet_v2,
    config={"bottle_neck": True, "filter_list": [64, 256, 512, 1024, 2048], "units": [3, 4, 23, 3]},
)
registry.register_model_config(
    "se_resnet_v2_152",
    SE_ResNet_v2,
    config={"bottle_neck": True, "filter_list": [64, 256, 512, 1024, 2048], "units": [3, 8, 36, 3]},
)
registry.register_model_config(
    "se_resnet_v2_200",
    SE_ResNet_v2,
    config={"bottle_neck": True, "filter_list": [64, 256, 512, 1024, 2048], "units": [3, 24, 36, 3]},
)
registry.register_model_config(
    "se_resnet_v2_269",
    SE_ResNet_v2,
    config={"bottle_neck": True, "filter_list": [64, 256, 512, 1024, 2048], "units": [3, 30, 48, 8]},
)



"""
CaiT, adapted from
https://github.com/facebookresearch/deit/blob/main/cait_models.py
and
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/cait.py

"Going deeper with Image Transformers", https://arxiv.org/abs/2103.17239
"""

# Reference license: Apache-2.0 and Apache-2.0

from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import StochasticDepth

from birder.model_registry import registry
from birder.net.base import BaseNet
from birder.net.vit import adjust_position_embedding


class PatchEmbed(nn.Module):
    def __init__(self, in_channels: int, embed_dim: int, patch_size: tuple[int, int]) -> None:
        super().__init__()
        self.proj = nn.Conv2d(in_channels, embed_dim, kernel_size=patch_size, stride=patch_size, padding=(0, 0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.proj(x)
        x = x.flatten(2).transpose(1, 2)  # NCHW -> NLC

        return x


class ClassAttention(nn.Module):
    def __init__(self, dim: int, num_heads: int, qkv_bias: bool, proj_drop: float) -> None:
        super().__init__()
        self.num_heads = num_heads

        self.q = nn.Linear(dim, dim, bias=qkv_bias)
        self.k = nn.Linear(dim, dim, bias=qkv_bias)
        self.v = nn.Linear(dim, dim, bias=qkv_bias)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, N, C) = x.shape
        q = self.q(x[:, 0]).unsqueeze(1).reshape(B, 1, self.num_heads, C // self.num_heads).permute(0, 2, 1, 3)
        k = self.k(x).reshape(B, N, self.num_heads, C // self.num_heads).permute(0, 2, 1, 3)
        v = self.v(x).reshape(B, N, self.num_heads, C // self.num_heads).permute(0, 2, 1, 3)

        x_cls = F.scaled_dot_product_attention(q, k, v, dropout_p=0.0)  # pylint: disable=not-callable

        x_cls = x_cls.transpose(1, 2).reshape(B, 1, C)
        x_cls = self.proj(x_cls)
        x_cls = self.proj_drop(x_cls)

        return x_cls


class ClassAttentionBlock(nn.Module):
    def __init__(
        self, dim: int, num_heads: int, mlp_ratio: float, qkv_bias: bool, proj_drop: float, drop_path: float, eta: float
    ) -> None:
        super().__init__()
        self.norm1 = nn.LayerNorm(dim)

        self.attn = ClassAttention(dim, num_heads=num_heads, qkv_bias=qkv_bias, proj_drop=proj_drop)

        self.drop_path = StochasticDepth(drop_path, mode="row")
        self.norm2 = nn.LayerNorm(dim)
        self.mlp = MLP(dim, [int(dim * mlp_ratio), dim], activation_layer=nn.GELU, dropout=proj_drop)

        self.gamma1 = nn.Parameter(eta * torch.ones(dim))
        self.gamma2 = nn.Parameter(eta * torch.ones(dim))

    def forward(self, x: torch.Tensor, x_cls: torch.Tensor) -> torch.Tensor:
        u = torch.concat((x_cls, x), dim=1)
        x_cls = x_cls + self.drop_path(self.gamma1 * self.attn(self.norm1(u)))
        x_cls = x_cls + self.drop_path(self.gamma2 * self.mlp(self.norm2(x_cls)))

        return x_cls


class TalkingHeadAttn(nn.Module):
    def __init__(self, dim: int, num_heads: int, qkv_bias: bool, attn_drop: float, proj_drop: float) -> None:
        super().__init__()

        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)

        self.proj = nn.Linear(dim, dim)
        self.proj_l = nn.Linear(num_heads, num_heads)
        self.proj_w = nn.Linear(num_heads, num_heads)

        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, N, C) = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q = qkv[0] * self.scale
        k = qkv[1]
        v = qkv[2]

        attn = q @ k.transpose(-2, -1)
        attn = self.proj_l(attn.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)
        attn = attn.softmax(dim=-1)
        attn = self.proj_w(attn.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class LayerScaleBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        mlp_ratio: float,
        qkv_bias: bool,
        proj_drop: float,
        attn_drop: float,
        drop_path: float,
        init_values: float,
    ) -> None:
        super().__init__()
        self.norm1 = nn.LayerNorm(dim)
        self.attn = TalkingHeadAttn(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            attn_drop=attn_drop,
            proj_drop=proj_drop,
        )
        self.drop_path = StochasticDepth(drop_path, mode="row")
        self.norm2 = nn.LayerNorm(dim)
        self.mlp = MLP(dim, [int(dim * mlp_ratio), dim], activation_layer=nn.GELU, dropout=proj_drop)
        self.gamma_1 = nn.Parameter(init_values * torch.ones(dim))
        self.gamma_2 = nn.Parameter(init_values * torch.ones(dim))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.drop_path(self.gamma_1 * self.attn(self.norm1(x)))
        x = x + self.drop_path(self.gamma_2 * self.mlp(self.norm2(x)))
        return x


class CaiT(BaseNet):
    block_group_regex = r"block1\.(\d+)"  # ClassAttentionBlock combined with the head

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        proj_drop = 0.0
        attn_drop = 0.0
        cls_attn_layers = 2
        mlp_ratio = 4.0
        qkv_bias = True
        patch_size: tuple[int, int] = self.config.get("patch_size", (16, 16))
        embed_dim: int = self.config["embed_dim"]
        depth: int = self.config["depth"]
        num_heads: int = self.config["num_heads"]
        init_values: float = self.config["init_values"]
        drop_path_rate: float = self.config["drop_path_rate"]

        self.patch_size = patch_size
        self.patch_embed = PatchEmbed(self.input_channels, embed_dim, patch_size)
        num_patches = (self.size[0] // patch_size[0]) * (self.size[1] // patch_size[1])

        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.pos_embed = nn.Parameter(torch.zeros(1, num_patches, embed_dim))

        dpr = [drop_path_rate for _ in range(depth)]  # Uniform stochastic depth

        layers1 = []
        for i in range(depth):
            layers1.append(
                LayerScaleBlock(
                    dim=embed_dim,
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    proj_drop=proj_drop,
                    attn_drop=attn_drop,
                    drop_path=dpr[i],
                    init_values=init_values,
                )
            )

        self.block1 = nn.Sequential(*layers1)

        self.block2 = nn.ModuleList()
        for _ in range(cls_attn_layers):
            self.block2.append(
                ClassAttentionBlock(
                    dim=embed_dim,
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    proj_drop=proj_drop,
                    drop_path=0.0,
                    eta=init_values,
                )
            )

        self.norm = nn.LayerNorm(embed_dim)

        self.embedding_size = embed_dim
        self.classifier = self.create_classifier()

        # Weights initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

            elif isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.bias, 0)
                nn.init.constant_(m.weight, 1.0)

        nn.init.trunc_normal_(self.pos_embed, std=0.02)
        nn.init.trunc_normal_(self.cls_token, std=0.02)

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.patch_embed(x)
        x = x + self.pos_embed
        x = self.block1(x)
        cls_tokens = self.cls_token.expand(x.shape[0], -1, -1)
        for blk in self.block2:
            cls_tokens = blk(x, cls_tokens)

        x = torch.concat((cls_tokens, x), dim=1)
        x = self.norm(x)

        return x

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return x[:, 0]

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        assert dynamic_size is False, "Dynamic size not supported for this network"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        old_size = self.size
        super().adjust_size(new_size)

        # Add back class tokens
        self.pos_embed = nn.Parameter(
            adjust_position_embedding(
                self.pos_embed,
                (old_size[0] // self.patch_size[0], old_size[1] // self.patch_size[1]),
                (new_size[0] // self.patch_size[0], new_size[1] // self.patch_size[1]),
                0,
            )
        )


registry.register_model_config(
    "cait_xxs24",
    CaiT,
    config={"embed_dim": 192, "depth": 24, "num_heads": 4, "init_values": 1e-5, "drop_path_rate": 0.05},
)
registry.register_model_config(
    "cait_xxs36",
    CaiT,
    config={"embed_dim": 192, "depth": 36, "num_heads": 4, "init_values": 1e-5, "drop_path_rate": 0.1},
)
registry.register_model_config(
    "cait_xs24",
    CaiT,
    config={"embed_dim": 288, "depth": 24, "num_heads": 6, "init_values": 1e-5, "drop_path_rate": 0.05},
)
registry.register_model_config(
    "cait_s24",
    CaiT,
    config={"embed_dim": 384, "depth": 24, "num_heads": 8, "init_values": 1e-5, "drop_path_rate": 0.1},
)
registry.register_model_config(
    "cait_s36",
    CaiT,
    config={"embed_dim": 384, "depth": 36, "num_heads": 8, "init_values": 1e-6, "drop_path_rate": 0.2},
)
registry.register_model_config(
    "cait_m36",
    CaiT,
    config={"embed_dim": 768, "depth": 36, "num_heads": 16, "init_values": 1e-6, "drop_path_rate": 0.3},
)
registry.register_model_config(
    "cait_m48",
    CaiT,
    config={"embed_dim": 768, "depth": 48, "num_heads": 16, "init_values": 1e-6, "drop_path_rate": 0.4},
)

registry.register_weights(
    "cait_xxs24_il-common",
    {
        "description": "CaiT XXS-24 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 45.4,
                "sha256": "1e7b3ba14b34b3d849eb0c86720e15804631e91ed51c74352c96e289583b5883",
            }
        },
        "net": {"network": "cait_xxs24", "tag": "il-common"},
    },
)



"""
MnasNet (B variant), adapted from
https://github.com/pytorch/vision/blob/main/torchvision/models/mnasnet.py

Paper "MnasNet: Platform-Aware Neural Architecture Search for Mobile", https://arxiv.org/abs/1807.11626

Changes from original:
* Relaxed the paper suggestion of 0.9997 momentum (1.0 - 0.9997 for Pytorch), using 0.99
"""

# Reference license: BSD 3-Clause

from collections import OrderedDict
from collections.abc import Callable
from functools import partial
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


def _round_to_multiple_of(val: float, divisor: int, round_up_bias: float = 0.9) -> int:
    """
    Asymmetric rounding to make val divisible by divisor.
    """

    if 0.0 >= round_up_bias or round_up_bias >= 1.0:
        raise ValueError(f"round_up_bias should be greater than 0.0 and smaller than 1.0 instead of {round_up_bias}")

    new_val = max(divisor, int(val + divisor / 2) // divisor * divisor)
    if new_val >= round_up_bias * val:
        return new_val

    return new_val + divisor


def _get_depths(alpha: float) -> list[int]:
    """
    Scales tensor depths as in reference MobileNet code, prefers rounding up rather than down
    """

    depths = [32, 16, 24, 40, 80, 96, 192, 320]
    return [_round_to_multiple_of(depth * alpha, 8) for depth in depths]


class InvertedResidual(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        padding: tuple[int, int],
        expansion_factor: float,
        norm_layer: Callable[..., nn.Module],
    ) -> None:
        super().__init__()
        mid_channels = in_channels * expansion_factor
        if in_channels == out_channels and stride[0] == 1:
            self.shortcut = True
        else:
            self.shortcut = False

        self.block = nn.Sequential(
            Conv2dNormActivation(
                in_channels,
                mid_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
                norm_layer=norm_layer,
            ),
            Conv2dNormActivation(
                mid_channels,
                mid_channels,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                groups=mid_channels,
                bias=False,
                norm_layer=norm_layer,
            ),
            Conv2dNormActivation(
                mid_channels,
                out_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
                norm_layer=norm_layer,
                activation_layer=None,
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.shortcut is True:
            return x + self.block(x)

        return self.block(x)


class InvertedResidualBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        padding: tuple[int, int],
        expansion_factor: float,
        repeats: int,
        norm_layer: Callable[..., nn.Module],
    ) -> None:
        super().__init__()
        layers = []
        layers.append(
            InvertedResidual(in_channels, out_channels, kernel_size, stride, padding, expansion_factor, norm_layer)
        )

        for _ in range(1, repeats):
            layers.append(
                InvertedResidual(out_channels, out_channels, kernel_size, (1, 1), padding, expansion_factor, norm_layer)
            )

        self.layers = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.layers(x)


class MNASNet(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        alpha: float = self.config["alpha"]

        depths = _get_depths(alpha)
        norm_layer = partial(nn.BatchNorm2d, momentum=1 - 0.99)

        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels,
                depths[0],
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                bias=False,
                norm_layer=norm_layer,
            ),
            Conv2dNormActivation(
                depths[0],
                depths[0],
                kernel_size=(3, 3),
                stride=(1, 1),
                padding=(1, 1),
                groups=depths[0],
                bias=False,
                norm_layer=norm_layer,
            ),
            Conv2dNormActivation(
                depths[0],
                depths[1],
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
                norm_layer=norm_layer,
                activation_layer=None,
            ),
        )

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        stages["stage1"] = nn.Sequential(
            InvertedResidualBlock(depths[1], depths[2], (3, 3), (2, 2), (1, 1), 3, 3, norm_layer),
        )
        return_channels.append(depths[2])
        stages["stage2"] = nn.Sequential(
            InvertedResidualBlock(depths[2], depths[3], (5, 5), (2, 2), (2, 2), 3, 3, norm_layer),
        )
        return_channels.append(depths[3])
        stages["stage3"] = nn.Sequential(
            InvertedResidualBlock(depths[3], depths[4], (5, 5), (2, 2), (2, 2), 6, 3, norm_layer),
            InvertedResidualBlock(depths[4], depths[5], (3, 3), (1, 1), (1, 1), 6, 2, norm_layer),
        )
        return_channels.append(depths[5])
        stages["stage4"] = nn.Sequential(
            InvertedResidualBlock(depths[5], depths[6], (5, 5), (2, 2), (2, 2), 6, 4, norm_layer),
            InvertedResidualBlock(depths[6], depths[7], (3, 3), (1, 1), (1, 1), 6, 1, norm_layer),
        )
        return_channels.append(depths[7])
        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            Conv2dNormActivation(
                depths[7],
                1280,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
                norm_layer=norm_layer,
            ),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
            nn.Dropout(p=0.2),
        )
        self.return_channels = return_channels
        self.embedding_size = 1280
        self.classifier = self.create_classifier()

        # Weights initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

            elif isinstance(m, nn.Linear):
                nn.init.kaiming_uniform_(m.weight, mode="fan_out", nonlinearity="sigmoid")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config("mnasnet_0_5", MNASNet, config={"alpha": 0.5})
registry.register_model_config("mnasnet_0_75", MNASNet, config={"alpha": 0.75})
registry.register_model_config("mnasnet_1_0", MNASNet, config={"alpha": 1.0})
registry.register_model_config("mnasnet_1_3", MNASNet, config={"alpha": 1.3})

registry.register_weights(
    "mnasnet_0_5_il-common",
    {
        "description": "MnasNet with depth multiplier of 0.5 trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 5.6,
                "sha256": "bec1d292d6400b54958206e9f89ec50a31e42abd94447fb466161dd0623cd75a",
            }
        },
        "net": {"network": "mnasnet_0_5", "tag": "il-common"},
    },
)
registry.register_weights(
    "mnasnet_1_0_il-common",
    {
        "description": "MnasNet with depth multiplier of 1 trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 13.9,
                "sha256": "351256cf362e5ed64838fb72e1455aee05793d2eff3c2c02f5a4c0f89cce7e53",
            }
        },
        "net": {"network": "mnasnet_1_0", "tag": "il-common"},
    },
)
